# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from sktime.transformations.base import _PanelToPanelTransformer
from sktime.utils.validation.panel import check_X

__all__ = ["TruncationTransformer"]
__author__ = ["Aaron Bostrom"]


class TruncationTransformer(_PanelToPanelTransformer):
    """TruncationTransformer docstring

    Parameters
    ----------
    lower   : int, optional (default=None) bottom range of the values to
                truncate can also be used to truncate to a specific length

                if None, will find the shortest sequence and use instead.

    upper   : int, optional (default=None) upper range, only required when
                paired with lower.
                This is used to calculate the range between. exclusive.

                if None, will truncate from 0 to the lower bound.
    """

    def __init__(self, lower=None, upper=None):
        self.lower = lower
        self.upper = upper
        self.min_length = lower
        super(TruncationTransformer, self).__init__()

    @staticmethod
    def get_min_length(X):
        def get_length(input):
            return min(map(lambda series: len(series), input))

        return min(map(get_length, X))

    def fit(self, X, y=None):
        """
        Fit transformer.

        Parameters
        ----------
        X : pandas DataFrame of shape [n_samples, n_features]
            Input data
        y : pandas Series, shape (n_samples, ...), optional
            Targets for supervised learning.

        Returns
        -------
        self : an instance of self.
        """
        X = check_X(X, coerce_to_pandas=True)

        if self.lower is None:
            n_instances, _ = X.shape
            arr = [X.iloc[i, :].values for i in range(n_instances)]
            self.lower_ = self.get_min_length(arr)
        else:
            self.lower_ = self.lower

        self._is_fitted = True
        return self

    def transform(self, X, y=None):
        """
        Transform X.

        Parameters
        ----------
        X : nested pandas DataFrame of shape [n_instances, n_columns]
            Nested dataframe with time-series in cells.

        Returns
        -------
        Xt : pandas DataFrame
        """
        self.check_is_fitted()
        X = check_X(X, coerce_to_pandas=True)

        n_instances, _ = X.shape

        arr = [X.iloc[i, :].values for i in range(n_instances)]

        min_length = self.get_min_length(arr)

        if min_length < self.lower_:
            raise ValueError(
                "Error: min_length of series \
                    is less than the one found when fit or set."
            )

        # depending on inputs either find the shortest truncation.
        # or use the bounds.
        if self.upper is None:
            idxs = np.arange(self.lower_)
        else:
            idxs = np.arange(self.lower_, self.upper)

        truncate = [pd.Series([series.iloc[idxs] for series in out]) for out in arr]

        return pd.DataFrame(truncate)



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

# test API provided through BaseForecaster

__author__ = ["Markus Löning"]
__all__ = [
    "test_raises_not_fitted_error",
    "test_score",
    "test_predict_time_index",
    "test_update_predict_predicted_index",
    "test_update_predict_predicted_index_update_params",
    "test_y_multivariate_raises_error",
    "test_get_fitted_params",
    "test_predict_time_index_in_sample_full",
    "test_predict_pred_interval",
    "test_update_predict_single",
    "test_y_invalid_type_raises_error",
    "test_predict_time_index_with_X",
    "test_X_invalid_type_raises_error",
]

import numpy as np
import pandas as pd
import pytest

from sktime.exceptions import NotFittedError
from sktime.forecasting.model_selection import SlidingWindowSplitter
from sktime.forecasting.model_selection import temporal_train_test_split
from sktime.forecasting.tests._config import TEST_ALPHAS
from sktime.forecasting.tests._config import TEST_FHS
from sktime.forecasting.tests._config import TEST_OOS_FHS
from sktime.forecasting.tests._config import TEST_STEP_LENGTHS
from sktime.forecasting.tests._config import TEST_WINDOW_LENGTHS
from sktime.forecasting.tests._config import VALID_INDEX_FH_COMBINATIONS
from sktime.performance_metrics.forecasting import (
    mean_absolute_percentage_error,
)
from sktime.utils import all_estimators
from sktime.utils._testing.estimator_checks import _construct_instance
from sktime.utils._testing.forecasting import _assert_correct_pred_time_index
from sktime.utils._testing.forecasting import _get_expected_index_for_update_predict
from sktime.utils._testing.forecasting import _make_fh
from sktime.utils._testing.forecasting import make_forecasting_problem
from sktime.utils._testing.series import _make_series
from sktime.utils.validation.forecasting import check_fh

# get all forecasters
FORECASTERS = all_estimators(estimator_types="forecaster", return_names=False)
FH0 = 1
INVALID_INPUT_TYPES = [np.empty(20), list(), tuple()]

# testing data
y = make_forecasting_problem()
y_train, y_test = temporal_train_test_split(y, train_size=0.75)


@pytest.mark.parametrize("Forecaster", FORECASTERS)
def test_get_fitted_params(Forecaster):
    f = _construct_instance(Forecaster)
    f.fit(y_train, fh=FH0)
    try:
        params = f.get_fitted_params()
        assert isinstance(params, dict)

    except NotImplementedError:
        pass


@pytest.mark.parametrize("Forecaster", FORECASTERS)
def test_raises_not_fitted_error(Forecaster):
    # We here check extra method of the forecaster API: update and update_predict.
    f = _construct_instance(Forecaster)

    # predict is check in test suite for all estimators
    with pytest.raises(NotFittedError):
        f.update(y_test, update_params=False)

    with pytest.raises(NotFittedError):
        cv = SlidingWindowSplitter(fh=1, window_length=1, start_with_window=False)
        f.update_predict(y_test, cv=cv)

    try:
        with pytest.raises(NotFittedError):
            f.get_fitted_params()
    except NotImplementedError:
        pass


@pytest.mark.parametrize("Forecaster", FORECASTERS)
def test_y_multivariate_raises_error(Forecaster):
    # Check that multivariate y raises an appropriate error message.
    y = _make_series(n_columns=2)
    f = _construct_instance(Forecaster)
    with pytest.raises(ValueError, match=r"univariate"):
        f.fit(y, fh=FH0)


@pytest.mark.parametrize("Forecaster", FORECASTERS)
@pytest.mark.parametrize("y", INVALID_INPUT_TYPES)
def test_y_invalid_type_raises_error(Forecaster, y):
    f = _construct_instance(Forecaster)
    with pytest.raises(TypeError, match=r"type"):
        f.fit(y, fh=FH0)


@pytest.mark.parametrize("Forecaster", FORECASTERS)
@pytest.mark.parametrize("X", INVALID_INPUT_TYPES)
def test_X_invalid_type_raises_error(Forecaster, X):
    f = _construct_instance(Forecaster)
    try:
        with pytest.raises(TypeError, match=r"type"):
            f.fit(y_train, X, fh=FH0)
    except NotImplementedError as e:
        msg = str(e).lower()
        assert "exogenous" in msg


@pytest.mark.parametrize("Forecaster", FORECASTERS)
@pytest.mark.parametrize(
    "index_type, fh_type, is_relative", VALID_INDEX_FH_COMBINATIONS
)
@pytest.mark.parametrize("steps", TEST_FHS)  # fh steps
def test_predict_time_index(Forecaster, index_type, fh_type, is_relative, steps):
    # Check that predicted time index matches forecasting horizon.
    y_train = make_forecasting_problem(index_type=index_type)
    cutoff = y_train.index[-1]
    fh = _make_fh(cutoff, steps, fh_type, is_relative)
    f = _construct_instance(Forecaster)

    # Some estimators may not support all time index types and fh types, hence we
    # need to catch NotImplementedErrors.
    try:
        f.fit(y_train, fh=fh)
        y_pred = f.predict()
        _assert_correct_pred_time_index(y_pred.index, y_train.index[-1], fh)
    except NotImplementedError:
        pass


@pytest.mark.parametrize("Forecaster", FORECASTERS)
@pytest.mark.parametrize(
    "index_type, fh_type, is_relative", VALID_INDEX_FH_COMBINATIONS
)
@pytest.mark.parametrize("steps", TEST_OOS_FHS)  # fh steps
def test_predict_time_index_with_X(Forecaster, index_type, fh_type, is_relative, steps):
    # Check that predicted time index matches forecasting horizon.
    y, X = make_forecasting_problem(index_type=index_type, make_X=True)
    cutoff = y.index[len(y) // 2]
    fh = _make_fh(cutoff, steps, fh_type, is_relative)

    y_train, y_test, X_train, X_test = temporal_train_test_split(y, X, fh=fh)

    f = _construct_instance(Forecaster)
    # Some estimators may not support all time index types and fh types, hence we
    # need to catch NotImplementedErrors.
    try:
        f.fit(y_train, X_train, fh=fh)
        y_pred = f.predict(X=X_test)
        _assert_correct_pred_time_index(y_pred.index, y_train.index[-1], fh)
    except NotImplementedError:
        pass


@pytest.mark.parametrize("Forecaster", FORECASTERS)
@pytest.mark.parametrize(
    "index_type, fh_type, is_relative", VALID_INDEX_FH_COMBINATIONS
)
def test_predict_time_index_in_sample_full(
    Forecaster, index_type, fh_type, is_relative
):
    # Check that predicted time index matched forecasting horizon for full in-sample
    # predictions.
    y_train = make_forecasting_problem(index_type=index_type)
    cutoff = y_train.index[-1]
    steps = -np.arange(len(y_train))  # full in-sample fh
    fh = _make_fh(cutoff, steps, fh_type, is_relative)
    f = _construct_instance(Forecaster)
    # Some estimators may not support all time index types and fh types, hence we
    # need to catch NotImplementedErrors.
    try:
        f.fit(y_train, fh=fh)
        y_pred = f.predict()
        _assert_correct_pred_time_index(y_pred.index, y_train.index[-1], fh)
    except NotImplementedError:
        pass


def _check_pred_ints(pred_ints: list, y_train: pd.Series, y_pred: pd.Series, fh):
    # make iterable
    if isinstance(pred_ints, pd.DataFrame):
        pred_ints = [pred_ints]

    for pred_int in pred_ints:
        # check column naming convention
        assert list(pred_int.columns) == ["lower", "upper"]

        # check time index
        _assert_correct_pred_time_index(pred_int.index, y_train.index[-1], fh)

        # check values
        assert np.all(pred_int["upper"] > y_pred)
        assert np.all(pred_int["lower"] < y_pred)

        # check if errors are weakly monotonically increasing
        # pred_errors = y_pred - pred_int["lower"]
        # # assert pred_errors.is_mononotic_increasing
        # assert np.all(
        #     pred_errors.values[1:].round(4) >= pred_errors.values[:-1].round(4)
        # )


@pytest.mark.parametrize("Forecaster", FORECASTERS)
@pytest.mark.parametrize("fh", TEST_OOS_FHS)
@pytest.mark.parametrize("alpha", TEST_ALPHAS)
def test_predict_pred_interval(Forecaster, fh, alpha):
    # Check prediction intervals.
    f = _construct_instance(Forecaster)
    f.fit(y_train, fh=fh)
    try:
        y_pred, pred_ints = f.predict(return_pred_int=True, alpha=alpha)
        _check_pred_ints(pred_ints, y_train, y_pred, fh)

    except NotImplementedError:
        pass


@pytest.mark.parametrize("Forecaster", FORECASTERS)
@pytest.mark.parametrize("fh", TEST_OOS_FHS)
def test_score(Forecaster, fh):
    # Check score method
    f = _construct_instance(Forecaster)
    f.fit(y_train, fh=fh)
    y_pred = f.predict()

    fh_idx = check_fh(fh).to_indexer()  # get zero based index
    expected = mean_absolute_percentage_error(
        y_pred, y_test.iloc[fh_idx], symmetric=True
    )

    # compare with actual score
    f = _construct_instance(Forecaster)
    f.fit(y_train, fh=fh)
    actual = f.score(y_test.iloc[fh_idx], fh=fh)
    assert actual == expected


@pytest.mark.parametrize("Forecaster", FORECASTERS)
@pytest.mark.parametrize("fh", TEST_OOS_FHS)
@pytest.mark.parametrize("update_params", [True, False])
def test_update_predict_single(Forecaster, fh, update_params):
    # Check correct time index of update-predict
    f = _construct_instance(Forecaster)
    f.fit(y_train, fh=fh)
    y_pred = f.update_predict_single(y_test, update_params=update_params)
    _assert_correct_pred_time_index(y_pred.index, y_test.index[-1], fh)


def _check_update_predict_predicted_index(
    Forecaster, fh, window_length, step_length, update_params
):
    y = make_forecasting_problem(all_positive=True, index_type="datetime")
    y_train, y_test = temporal_train_test_split(y)
    cv = SlidingWindowSplitter(
        fh,
        window_length=window_length,
        step_length=step_length,
        start_with_window=False,
    )
    f = _construct_instance(Forecaster)
    f.fit(y_train, fh=fh)
    y_pred = f.update_predict(y_test, cv=cv, update_params=update_params)
    assert isinstance(y_pred, (pd.Series, pd.DataFrame))
    if isinstance(y_pred, pd.DataFrame):
        assert y_pred.shape[1] > 1
    expected = _get_expected_index_for_update_predict(y_test, fh, step_length)
    actual = y_pred.index
    np.testing.assert_array_equal(actual, expected)


# test with update_params=False and different values for steps_length
@pytest.mark.parametrize("Forecaster", FORECASTERS)
@pytest.mark.parametrize("fh", TEST_OOS_FHS)
@pytest.mark.parametrize("window_length", TEST_WINDOW_LENGTHS)
@pytest.mark.parametrize("step_length", TEST_STEP_LENGTHS)
@pytest.mark.parametrize("update_params", [False])
def test_update_predict_predicted_index(
    Forecaster, fh, window_length, step_length, update_params
):
    _check_update_predict_predicted_index(
        Forecaster, fh, window_length, step_length, update_params
    )


# test with update_params=True and step_length=1
@pytest.mark.parametrize("Forecaster", FORECASTERS)
@pytest.mark.parametrize("fh", TEST_OOS_FHS)
@pytest.mark.parametrize("window_length", TEST_WINDOW_LENGTHS)
@pytest.mark.parametrize("step_length", [1])
@pytest.mark.parametrize("update_params", [True])
def test_update_predict_predicted_index_update_params(
    Forecaster, fh, window_length, step_length, update_params
):
    _check_update_predict_predicted_index(
        Forecaster, fh, window_length, step_length, update_params
    )



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
"""copyright: sktime developers, BSD-3-Clause License (see LICENSE file)."""

__author__ = ["Guzal Bulatova"]
__all__ = []

import numpy as np
import pytest
from scipy.stats import linregress

from sktime.datasets import load_airline
from sktime.transformations.series.theta import ThetaLinesTransformer


def test_theta_0():
    # with theta = 0
    y = load_airline()
    t = ThetaLinesTransformer(0)
    t.fit(y)
    actual = t.transform(y)
    x = np.arange(y.size) + 1
    lin_regress = linregress(x, y)
    expected = lin_regress.intercept + lin_regress.slope * x

    np.testing.assert_almost_equal(actual, expected, decimal=8)


def test_theta_1():
    # with theta = 1 Theta-line is equal to the original time-series
    y = load_airline()
    t = ThetaLinesTransformer(1)
    t.fit(y)
    actual = t.transform(y)
    np.testing.assert_array_equal(actual, y)


@pytest.mark.parametrize("theta", [(1, 1.5), (0, 1, 2), (0.25, 0.5, 0.75, 1, 2)])
def test_thetalines_shape(theta):
    y = load_airline()
    t = ThetaLinesTransformer(theta)
    t.fit(y)
    actual = t.transform(y)
    assert actual.shape == (y.shape[0], len(theta))



# -*- coding: utf-8 -*-
"""
Lists of datasets available from the archive on timeseriesclassification.com

There are four main distinctions: univariate/multivariate equal/unequal length.
Array univariate lists the 128 UCR problems, as described in [1].
Array multivariate lists the 30 UEA problems, as described in [2], plus 3 new additions.
Array univariate_equal_length lists the 112 UCR archive problems used in [3].
Array multivariate_equal_length lists the 26 UEA archive problems used in [4].

[1] H.Dau, A. Bagnall, K. Kamgar, C. Yeh, Y. Zhu, S. Gharghabi, C. Ratanamahatana and
E. Keogh.
The  UCR  time  series  archive. IEEE/CAA J. Autom. Sinica, 6(6):1293–1305, 2019
[2] A. Bagnall, H. Dau, J. Lines, M. Flynn, J. Large, A. Bostrom, P. Southam,and
E.  Keogh.
The UEA  multivariate  time  series  classification  archive, 2018. ArXiv e-prints,
arXiv:1811.00075, 2018
[3] A. Bagnall, M. Flynn, J. Large, J. Lines and M. Middlehurst.
On the Usage and Performance of the Hierarchical Vote Collective of Transformation-Based
Ensembles Version 1.0
(HIVE-COTE v1.0). Lecture Notes in Computer Science. in proc. 5th Advanced Analytics and
Learning on Temporal Data
[4] A. Pasos Ruiz, M. Flynn, J. Large, M. Middlehurst and A. Bagnall.
    The great multivariate time series classification bake off: a review and
    experimental evaluation of recent algorithmic advances,
    Data Mining and Knowledge Discovery, 2020.
"""

__author__ = ["Tony Bagnall"]

""" 128 UCR univariatetime series classification problems [1]"""
univariate = [
    "ACSF1",
    "Adiac",
    "AllGestureWiimoteX",
    "AllGestureWiimoteY",
    "AllGestureWiimoteZ",
    "ArrowHead",
    "Beef",
    "BeetleFly",
    "BirdChicken",
    "BME",
    "Car",
    "CBF",
    "ChlorineConcentration",
    "CinCECGTorso",
    "Coffee",
    "Computers",
    "CricketX",
    "CricketY",
    "CricketZ",
    "Crop",
    "DiatomSizeReduction",
    "DistalPhalanxOutlineCorrect",
    "DistalPhalanxOutlineAgeGroup",
    "DistalPhalanxTW",
    "DodgerLoopDay",
    "DodgerLoopGame",
    "DodgerLoopWeekend",
    "Earthquakes",
    "ECG200",
    "ECG5000",
    "ECGFiveDays",
    "ElectricDevices",
    "EOGHorizontalSignal",
    "EOGVerticalSignal",
    "EthanolLevel",
    "FaceAll",
    "FaceFour",
    "FacesUCR",
    "FiftyWords",
    "Fish",
    "FordA",
    "FordB",
    "FreezerRegularTrain",
    "FreezerSmallTrain",
    "Fungi",
    "GestureMidAirD1",
    "GestureMidAirD2",
    "GestureMidAirD3",
    "GesturePebbleZ1",
    "GesturePebbleZ2",
    "GunPoint",
    "GunPointAgeSpan",
    "GunPointMaleVersusFemale",
    "GunPointOldVersusYoung",
    "Ham",
    "HandOutlines",
    "Haptics",
    "Herring",
    "HouseTwenty",
    "InlineSkate",
    "InsectEPGRegularTrain",
    "InsectEPGSmallTrain",
    "InsectWingbeatSound",
    "ItalyPowerDemand",
    "LargeKitchenAppliances",
    "Lightning2",
    "Lightning7",
    "Mallat",
    "Meat",
    "MedicalImages",
    "MelbournePedestrian",
    "MiddlePhalanxOutlineCorrect",
    "MiddlePhalanxOutlineAgeGroup",
    "MiddlePhalanxTW",
    "MixedShapesRegularTrain",
    "MixedShapesSmallTrain",
    "MoteStrain",
    "NonInvasiveFetalECGThorax1",
    "NonInvasiveFetalECGThorax2",
    "OliveOil",
    "OSULeaf",
    "PhalangesOutlinesCorrect",
    "Phoneme",
    "PickupGestureWiimoteZ",
    "PigAirwayPressure",
    "PigArtPressure",
    "PigCVP",
    "PLAID",
    "Plane",
    "PowerCons",
    "ProximalPhalanxOutlineCorrect",
    "ProximalPhalanxOutlineAgeGroup",
    "ProximalPhalanxTW",
    "RefrigerationDevices",
    "Rock",
    "ScreenType",
    "SemgHandGenderCh2",
    "SemgHandMovementCh2",
    "SemgHandSubjectCh2",
    "ShakeGestureWiimoteZ",
    "ShapeletSim",
    "ShapesAll",
    "SmallKitchenAppliances",
    "SmoothSubspace",
    "SonyAIBORobotSurface1",
    "SonyAIBORobotSurface2",
    "StarlightCurves",
    "Strawberry",
    "SwedishLeaf",
    "Symbols",
    "SyntheticControl",
    "ToeSegmentation1",
    "ToeSegmentation2",
    "Trace",
    "TwoLeadECG",
    "TwoPatterns",
    "UMD",
    "UWaveGestureLibraryAll",
    "UWaveGestureLibraryX",
    "UWaveGestureLibraryY",
    "UWaveGestureLibraryZ",
    "Wafer",
    "Wine",
    "WordSynonyms",
    "Worms",
    "WormsTwoClass",
    "Yoga",
]

""" 33 UEA multivariate time series classification problems [2]"""
multivariate = [
    "ArticularyWordRecognition",
    "AsphaltObstaclesCoordinates",
    "AsphaltPavementTypeCoordinates",
    "AsphaltRegularityCoordinates",
    "AtrialFibrillation",
    "BasicMotions",
    "CharacterTrajectories",
    "Cricket",
    "DuckDuckGeese",
    "EigenWorms",
    "Epilepsy",
    "EthanolConcentration",
    "ERing",
    "FaceDetection",
    "FingerMovements",
    "HandMovementDirection",
    "Handwriting",
    "Heartbeat",
    "InsectWingbeat",
    "JapaneseVowels",
    "Libras",
    "LSST",
    "MotorImagery",
    "NATOPS",
    "PenDigits",
    "PEMS-SF",
    "PhonemeSpectra",
    "RacketSports",
    "SelfRegulationSCP1",
    "SelfRegulationSCP2",
    "SpokenArabicDigits",
    "StandWalkJump",
    "UWaveGestureLibrary",
]

"""113 equal length/no missing univariate time series classification problems [3]"""
univariate_equal_length = [
    "ACSF1",
    "Adiac",
    "ArrowHead",
    "Beef",
    "BeetleFly",
    "BirdChicken",
    "BME",
    "Car",
    "CBF",
    "UnitTest",
    "ChlorineConcentration",
    "CinCECGTorso",
    "Coffee",
    "Computers",
    "CricketX",
    "CricketY",
    "CricketZ",
    "Crop",
    "DiatomSizeReduction",
    "DistalPhalanxOutlineCorrect",
    "DistalPhalanxOutlineAgeGroup",
    "DistalPhalanxTW",
    "Earthquakes",
    "ECG200",
    "ECG5000",
    "ECGFiveDays",
    "ElectricDevices",
    "EOGHorizontalSignal",
    "EOGVerticalSignal",
    "EthanolLevel",
    "FaceAll",
    "FaceFour",
    "FacesUCR",
    "FiftyWords",
    "Fish",
    "FordA",
    "FordB",
    "FreezerRegularTrain",
    "FreezerSmallTrain",
    "Fungi",
    "GunPoint",
    "GunPointAgeSpan",
    "GunPointMaleVersusFemale",
    "GunPointOldVersusYoung",
    "Ham",
    "HandOutlines",
    "Haptics",
    "Herring",
    "HouseTwenty",
    "InlineSkate",
    "InsectEPGRegularTrain",
    "InsectEPGSmallTrain",
    "InsectWingbeatSound",
    "ItalyPowerDemand",
    "LargeKitchenAppliances",
    "Lightning2",
    "Lightning7",
    "Mallat",
    "Meat",
    "MedicalImages",
    "MiddlePhalanxOutlineCorrect",
    "MiddlePhalanxOutlineAgeGroup",
    "MiddlePhalanxTW",
    "MixedShapesRegularTrain",
    "MixedShapesSmallTrain",
    "MoteStrain",
    "NonInvasiveFetalECGThorax1",
    "NonInvasiveFetalECGThorax2",
    "OliveOil",
    "OSULeaf",
    "PhalangesOutlinesCorrect",
    "Phoneme",
    "PigAirwayPressure",
    "PigArtPressure",
    "PigCVP",
    "Plane",
    "PowerCons",
    "ProximalPhalanxOutlineCorrect",
    "ProximalPhalanxOutlineAgeGroup",
    "ProximalPhalanxTW",
    "RefrigerationDevices",
    "Rock",
    "ScreenType",
    "SemgHandGenderCh2",
    "SemgHandMovementCh2",
    "SemgHandSubjectCh2",
    "ShapeletSim",
    "ShapesAll",
    "SmallKitchenAppliances",
    "SmoothSubspace",
    "SonyAIBORobotSurface1",
    "SonyAIBORobotSurface2",
    "StarlightCurves",
    "Strawberry",
    "SwedishLeaf",
    "Symbols",
    "SyntheticControl",
    "ToeSegmentation1",
    "ToeSegmentation2",
    "Trace",
    "TwoLeadECG",
    "TwoPatterns",
    "UMD",
    "UWaveGestureLibraryAll",
    "UWaveGestureLibraryX",
    "UWaveGestureLibraryY",
    "UWaveGestureLibraryZ",
    "Wafer",
    "Wine",
    "WordSynonyms",
    "Worms",
    "WormsTwoClass",
    "Yoga",
]

"""11 variable length univariate time series classification problems [3]"""
univariate_variable_length = [
    "AllGestureWiimoteX",
    "AllGestureWiimoteY",
    "AllGestureWiimoteZ",
    "GestureMidAirD1",
    "GestureMidAirD2",
    "GestureMidAirD3",
    "GesturePebbleZ1",
    "GesturePebbleZ2",
    "PickupGestureWiimoteZ",
    "PLAID",
    "ShakeGestureWiimoteZ",
]

"""4 fixed length univariate time series classification problems with missing values"""
univariate_missing_values = [
    "DodgerLoopDay",
    "DodgerLoopGame",
    "DodgerLoopWeekend",
    "MelbournePedestrian",
]

"""26 equal length multivariate time series classification problems [4]"""
multivariate_equal_length = [
    "ArticularyWordRecognition",
    "AtrialFibrillation",
    "BasicMotions",
    "Cricket",
    "DuckDuckGeese",
    "EigenWorms",
    "Epilepsy",
    "EthanolConcentration",
    "ERing",
    "FaceDetection",
    "FingerMovements",
    "HandMovementDirection",
    "Handwriting",
    "Heartbeat",
    "Libras",
    "LSST",
    "MotorImagery",
    "NATOPS",
    "PenDigits",
    "PEMS-SF",
    "PhonemeSpectra",
    "RacketSports",
    "SelfRegulationSCP1",
    "SelfRegulationSCP2",
    "StandWalkJump",
    "UWaveGestureLibrary",
]

"""7 variable length multivariate time series classification problems [4]"""
multivariate_unequal_length = [
    "AsphaltObstaclesCoordinates",
    "AsphaltPavementTypeCoordinates",
    "AsphaltRegularityCoordinates",
    "CharacterTrajectories",
    "InsectWingbeat",
    "JapaneseVowels",
    "SpokenArabicDigits",
]



# -*- coding: utf-8 -*-
__all__ = ["Orchestrator"]
__author__ = ["Viktor Kazakov", "Markus Löning"]

from sklearn.base import clone
from sktime.benchmarking.tasks import TSCTask
from sktime.benchmarking.tasks import TSRTask
import pandas as pd
import logging

log = logging.getLogger()
console = logging.StreamHandler()
log.addHandler(console)


class Orchestrator:
    """
    Fit and predict one or more estimators on one or more datasets
    """

    def __init__(self, tasks, datasets, strategies, cv, results):
        # validate datasets and tasks
        self._validate_tasks_and_datasets(tasks, datasets)
        self.tasks = tasks
        self.datasets = datasets

        # validate strategies
        self._validate_strategy_names(strategies)
        self.strategies = strategies

        self.cv = cv
        self.results = results

        # attach cv iterator to results object
        self.results.cv = cv

        # progress trackers
        self.n_strategies = len(strategies)
        self.n_datasets = len(datasets)
        self._strategy_counter = 0
        self._dataset_counter = 0

    def _iter(self):
        """Iterator for orchestration"""
        # TODO: check if datasets are skipped entirely because predictions
        #  already exists before loading data,
        #  maybe do a dry-run first to find out which datasets to skip?
        for task, dataset in zip(self.tasks, self.datasets):
            # update counters
            self._strategy_counter = 0
            self._dataset_counter += 1

            # load data into memory from dataset hook
            data = dataset.load()

            # get target in case stratified cross-validation is used
            y = data[task.target]

            for strategy in self.strategies:
                self._strategy_counter += 1  # update counter

                for cv_fold, (train_idx, test_idx) in enumerate(self.cv.split(data, y)):
                    # for each fold, clone strategy to avoid updating
                    # already fitted strategies
                    strategy = clone(strategy)

                    yield (task, dataset, data, strategy, cv_fold, train_idx, test_idx)

    def fit(self, overwrite_fitted_strategies=False, verbose=False):
        """Fit strategies on datasets"""

        for (
            task,
            dataset,
            data,
            strategy,
            cv_fold,
            train_idx,
            _test_idx,
        ) in self._iter():

            # skip strategy, if overwrite is set to False and fitted
            # strategy already exists
            if (
                not overwrite_fitted_strategies
                and self.results.check_fitted_strategy_exists(
                    strategy, data.dataset_name
                )
            ):
                log.warn(
                    f"Skipping strategy: {strategy.name} on CV-fold: "
                    f"{cv_fold} of dataset: {dataset.name}"
                )
                continue

            # else fit and save fitted strategy
            else:
                train = data.iloc[train_idx]
                self._print_progress(
                    dataset.name, strategy.name, cv_fold, "train", "fit", verbose
                )
                strategy.fit(task, train)
                self.results.save_fitted_strategy(
                    strategy=strategy, dataset_name=data.dataset_name, cv_fold=cv_fold
                )

    def predict(
        self, overwrite_predictions=False, predict_on_train=False, verbose=False
    ):
        """Predict from saved fitted strategies"""
        raise NotImplementedError(
            "Predicting from saved fitted strategies is not implemented yet"
        )

    def fit_predict(
        self,
        overwrite_predictions=False,
        predict_on_train=False,
        save_fitted_strategies=True,
        overwrite_fitted_strategies=False,
        verbose=False,
    ):
        """Fit and predict"""

        # check that for fitted strategies overwrite option is only set when
        # save option is set
        if overwrite_fitted_strategies and not save_fitted_strategies:
            raise ValueError(
                f"Can only overwrite fitted strategies "
                f"if fitted strategies are saved, but found: "
                f"overwrite_fitted_strategies="
                f"{overwrite_fitted_strategies} and "
                f"save_fitted_strategies="
                f"{save_fitted_strategies}"
            )

        # fitting and prediction
        for task, dataset, data, strategy, cv_fold, train_idx, test_idx in self._iter():

            # check which results already exist
            train_pred_exist = self.results.check_predictions_exist(
                strategy.name, dataset.name, cv_fold, train_or_test="train"
            )
            test_pred_exist = self.results.check_predictions_exist(
                strategy.name, dataset.name, cv_fold, train_or_test="test"
            )
            fitted_stategy_exists = self.results.check_fitted_strategy_exists(
                strategy.name, dataset.name, cv_fold
            )

            # skip if overwrite is set to False for both predictions and
            # strategies and all results exist
            if (
                not overwrite_predictions
                and test_pred_exist
                and (train_pred_exist or not predict_on_train)
                and not overwrite_fitted_strategies
                and (fitted_stategy_exists or not save_fitted_strategies)
            ):
                log.warn(
                    f"Skipping strategy: {strategy.name} on CV-fold: "
                    f"{cv_fold} of dataset: {dataset.name}"
                )
                continue

            # split data into training and test sets
            train = data.iloc[train_idx]
            test = data.iloc[test_idx]

            # fit strategy
            self._print_progress(
                dataset.name, strategy.name, cv_fold, "train", "fit", verbose
            )
            fit_estimator_start_time = pd.Timestamp.now()
            strategy.fit(task, train)
            fit_estimator_end_time = pd.Timestamp.now()

            # save fitted strategy if save fitted strategies is set to True
            # and overwrite is set to True or the
            # fitted strategy does not already exist
            if save_fitted_strategies and (
                overwrite_fitted_strategies or not fitted_stategy_exists
            ):
                self.results.save_fitted_strategy(
                    strategy, dataset_name=dataset.name, cv_fold=cv_fold
                )

            # optionally, predict on training set if predict on train is set
            # to True and and overwrite is set to True
            # or the predicted values do not already exist
            if predict_on_train and (overwrite_predictions or not train_pred_exist):
                y_true = train.loc[:, task.target]
                predict_estimator_start_time = pd.Timestamp.now()
                y_pred = strategy.predict(train)
                predict_estimator_end_time = pd.Timestamp.now()

                y_proba = self._predict_proba_one(strategy, task, train, y_true, y_pred)
                self.results.save_predictions(
                    strategy_name=strategy.name,
                    dataset_name=dataset.name,
                    index=train_idx,
                    y_true=y_true,
                    y_pred=y_pred,
                    y_proba=y_proba,
                    cv_fold=cv_fold,
                    fit_estimator_start_time=fit_estimator_start_time,
                    fit_estimator_end_time=fit_estimator_end_time,
                    predict_estimator_start_time=predict_estimator_start_time,
                    predict_estimator_end_time=predict_estimator_end_time,
                    train_or_test="train",
                )

            # predict on test set if overwrite predictions is set to True or
            # predictions do not already exist
            if overwrite_predictions or not test_pred_exist:
                y_true = test.loc[:, task.target]
                predict_estimator_start_time = pd.Timestamp.now()
                y_pred = strategy.predict(test)
                predict_estimator_end_time = pd.Timestamp.now()

                y_proba = self._predict_proba_one(strategy, task, test, y_true, y_pred)
                self.results.save_predictions(
                    dataset_name=dataset.name,
                    strategy_name=strategy.name,
                    index=test_idx,
                    y_true=y_true,
                    y_pred=y_pred,
                    y_proba=y_proba,
                    cv_fold=cv_fold,
                    fit_estimator_start_time=fit_estimator_start_time,
                    fit_estimator_end_time=fit_estimator_end_time,
                    predict_estimator_start_time=predict_estimator_start_time,
                    predict_estimator_end_time=predict_estimator_end_time,
                    train_or_test="test",
                )

        # save results as master file
        self.results.save()

    @staticmethod
    def _predict_proba_one(strategy, task, data, y_true, y_pred):
        """Predict strategy on one dataset"""
        # TODO always try to get probabilistic predictions first, compute
        #  deterministic predictions using
        #  argmax to avoid rerunning predictions, only if no predict_proba
        #  is available, run predict

        # if the task is classification and the strategies supports
        # probabilistic predictions,
        # get probabilistic predictions
        if isinstance(task, TSCTask) and hasattr(strategy, "predict_proba"):
            return strategy.predict_proba(data)

            # otherwise, return deterministic predictions in expected format
            # else:
            #     n_class_true = len(np.unique(y_true))
            #     n_class_pred = len(np.unique(y_pred))
            #     n_classes = np.maximum(n_class_pred, n_class_true)
            #     n_predictions = len(y_pred)
            #     y_proba = (n_predictions, n_classes)
            #     y_proba = np.zeros(y_proba)
            #     y_proba[:, np.array(y_pred, dtype=np.int)] = 1

        else:
            return None

    @staticmethod
    def _validate_strategy_names(strategies):
        """Validate strategy names"""

        # Check uniqueness of strategy names
        names = [strategy.name for strategy in strategies]
        if not len(names) == len(set(names)):
            raise ValueError(
                f"Names of provided strategies are not unique: " f"{names}"
            )

        # Check for conflicts with estimator kwargs
        all_params = []
        for strategy in strategies:
            params = list(strategy.get_params(deep=False).keys())
            all_params.extend(params)

        invalid_names = set(names).intersection(set(all_params))
        if invalid_names:
            raise ValueError(
                f"Strategy names conflict with constructor "
                f"arguments: {sorted(invalid_names)}"
            )

        # Check for conflicts with double-underscore convention
        invalid_names = [name for name in names if "__" in name]
        if invalid_names:
            raise ValueError(
                f"Estimator names must not contain __: got " f"{invalid_names}"
            )

    @staticmethod
    def _validate_tasks_and_datasets(tasks, datasets):
        """Validate tasks"""
        # check input types
        if not isinstance(datasets, list):
            raise ValueError(f"datasets must be a list, but found: {type(datasets)}")
        if not isinstance(tasks, list):
            raise ValueError(f"tasks must be a list, but found: {type(tasks)}")

        # check if there is one task for each dataset
        if len(tasks) != len(datasets):
            raise ValueError(
                "Inconsistent number of datasets and tasks, "
                "there must be one task for each dataset"
            )

        # check if task is either time series regression or classification,
        # other tasks not supported yet
        if not all(isinstance(task, (TSCTask, TSRTask)) for task in tasks):
            raise NotImplementedError(
                "Currently, only time series classification and time series "
                "regression tasks are supported"
            )

        # check if all tasks are of the same type
        if not all(isinstance(task, type(tasks[0])) for task in tasks):
            raise ValueError("Not all tasks are of the same type")

    def _print_progress(
        self,
        dataset_name,
        strategy_name,
        cv_fold,
        train_or_test,
        fit_or_predict,
        verbose,
    ):
        """Helper function to print progress"""

        if verbose:
            fit_or_predict = fit_or_predict.capitalize()
            if train_or_test == "train" and fit_or_predict == "predict":
                on_train = " (training set)"
            else:
                on_train = ""

            n_splits = self.cv.get_n_splits() - 1  # zero indexing

            log.warn(
                f"strategy: {self._strategy_counter}/{self.n_strategies} - "
                f"{strategy_name} "
                f"on CV-fold: {cv_fold}/{n_splits} "
                f"of dataset: {self._dataset_counter}/{self.n_datasets} - "
                f"{dataset_name}{on_train}"
            )



from sklearn.ensemble._forest import ForestClassifier
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble._forest import _generate_sample_indices, compute_sample_weight
import numpy as np
from numpy import random
from sklearn.decomposition import PCA
from scipy.stats import mode
from copy import deepcopy

#they extract random subsections of the data, and perform PCA with variancecoverage of 1.0
#the number of subsections of the data is based on the method,
#generateGroupsFromNumbers or generateGroupsFromSizes.

class RotationForest(ForestClassifier):

    def __init__(self,
                 n_estimators=10,
                 minGroup=3,
                 maxGroup=3,
                 remove_percentage=50,
                 random_state=None,
                 verbose=0,):
        super(RotationForest, self).__init__(
            base_estimator=DecisionTreeClassifier(),
            n_estimators=n_estimators)

        self.verbose = verbose
        self.random_state = random_state
        self.minGroup = minGroup
        self.maxGroup = maxGroup
        self.remove_percentage = remove_percentage
        self.base_projectionFilter = PCA()
        self.base_estimator = DecisionTreeClassifier()
        self._std = []
        self._mean = []
        self._noise = []
        self._classifiers = []
        self._num_classes = 0
        self._num_atts = 0
        self.classes_ = None

        self._groups = [None] * self.n_estimators
        self._pcas = [None] * self.n_estimators

        random.seed(random_state)


    def fit(self, X, y, sample_weight=None):

        n_samps, self._num_atts = X.shape
        self._num_classes = np.unique(y).shape[0]
        self.classes_ = list(set(y))
        print(self._num_classes)
        print(self._num_atts)

        #need to remove 0 variance attributes.


        # Compute mean, std and noise for z-normalisation
        self._std = np.std(X, axis=0)
        self._med = np.mean(X, axis=0)
        self._noise = [random.uniform(-0.000005, 0.000005) for p in range(0, X.shape[1])] #in case we have 0 std.

        # Normalisae the data
        X = (X - self._med) / (self._std + self._noise)

        classGrouping = self.createClassesGrouping(X,y)
        #print(np.array(classGrouping).shape)


        for i in range(0, self.n_estimators):
            #generate groups.
            self._groups[i] = self.generateGroupsFromSizes(X)
            self._pcas[i] = []

            #construct the slices to fit the PCAs too.
            for j, grp in enumerate(self._groups[i]):
                rot_j = []
                #randomly add the classes
                #with the randomly selected attributes.
                for index, selected_class in enumerate(self.selectClasses()):
                    if not selected_class:
                        continue
                    #we have to reshape the array in case it's a size 1, and gets squeezed.
                    for inst in classGrouping[index]:
                        rot_j.append(inst[grp])
                #randomly sample 50% of the indices.
                sample_ind = random.choice(len(rot_j), int((float(len(rot_j))/100.0) * self.remove_percentage), replace=False)
                rot_j = np.array(rot_j).reshape((-1, grp.shape[0]))
                #only sample if we have lots of instances.
                if sample_ind.shape[0] > 2:
                    rot_j = rot_j[sample_ind]

				##try to fit the PCA if it fails, remake it, and add 10 random data instances.
                while True:
				    #ignore err state on PCA because we account if it fails.
                    with np.errstate(divide='ignore',invalid='ignore'):
                        pca_data = deepcopy(self.base_projectionFilter).fit(rot_j)

                    self.addRandomInstance(rot_j, 10)
                    if not np.isnan(pca_data.explained_variance_ratio_).all():
                        break
                    rot_j = self.addRandomInstance(rot_j, 10)

                self._pcas[i].append(pca_data)

            #merge all the pca_transformed data into one instance and build a classifier on it.
            transformed_X = self.convertData(X, i)
            #with this transformed data, we need to train a tree.
            tree = deepcopy(self.base_estimator)
            tree.fit(transformed_X, y)
            self._classifiers.append(tree)

    def addRandomInstance(self, data, num_instances):
        random_instances = np.random.random_sample((num_instances,data.shape[1]))
        output = np.concatenate([data,random_instances])
        return output

    def convertData(self, X, i):
       return np.array([self.createRow(row,i) for row in X])

    #could project along the axis, but then i couldn't use it for new instances.
    def createRow(self, row, i):
        return np.concatenate([self._pcas[i][j].transform([row[grp]])[0] for j, grp in enumerate(self._groups[i])], axis=0)

    def createClassesGrouping(self, X, Y):
        return np.array([np.squeeze(X[np.argwhere(Y == i)]).reshape((-1,self._num_atts))  for i in self.classes_])

    def selectClasses(self):
        numSelected = 0
        selected = np.zeros(self._num_classes, dtype=np.bool)

        for i in range(0,self._num_classes):
            if random.choice([True,False]):
                selected[i] = True
                numSelected +=1

        if numSelected == 0:
            selected[random.choice(self._num_classes)] = True
        return selected

    def attributes_permuation(self, numAtts, classAtt):
        #arange is no inclusive
        vals = np.concatenate(( np.arange(0, classAtt, 1), np.arange(classAtt, numAtts)))
        return np.random.permutation(vals)

    def generateGroupsFromSizes(self, data):
        # generate one permuation, then do it for all classifiers.

        numGroupsOfSize = np.zeros(self.maxGroup - self.minGroup + 1)
        permutation = self.attributes_permuation(self._num_atts, self._num_classes)

        perm_length = permutation.shape[0]

        #select the size of each group
        numAttributes = 0
        numGroups = 0
        while numAttributes < perm_length:
            n = random.randint(numGroupsOfSize.shape[0])
            numGroupsOfSize[n] += 1
            numAttributes += self.minGroup + n
            numGroups += 1

        currentAttribute =0
        currentSize = 0

        Groups = []
        for j in range(0, numGroups):
            while numGroupsOfSize[currentSize] == 0:
                currentSize +=1
            numGroupsOfSize[currentSize] -=1

            n = self.minGroup + currentSize
            Groups.append(np.zeros(n))
            for k in range(0,n):
                if currentAttribute < perm_length:
                    Groups[j][k] = permutation[currentAttribute]
                else:
                    Groups[j][k] = permutation[random.randint(perm_length)]
                currentAttribute+=1

        return np.array(Groups, dtype=np.int)

    def predict(self, X):
        return [self.classes_[np.argmax(prob)] for prob in self.predict_proba(X)]

    def predict_proba(self, X):
        #need to normalise and remove atts
        self._std = np.std(X, axis=0)
        self._med = np.mean(X, axis=0)
        self._noise = [random.uniform(-0.000005, 0.000005) for p in range(0, X.shape[1])] #in case we have 0 std.

        # Normalise the data
        X = (X - self._med) / (self._std + self._noise)

        sums = np.zeros((X.shape[0],self._num_classes))
        for i, clf in enumerate(self._classifiers):
            transformX = self.convertData(X,i)
            sums += clf.predict_proba(transformX)

        output = sums / (np.ones(self._num_classes) * self.n_estimators)
        return output



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import pytest
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import FunctionTransformer

from sktime.series_as_features.compose import FeatureUnion
from sktime.transformations.panel.compose import (
    SeriesToPrimitivesRowTransformer,
)
from sktime.transformations.panel.segment import RandomIntervalSegmenter
from sktime.transformations.panel.summarize import (
    RandomIntervalFeatureExtractor,
)
from sktime.utils._testing.panel import make_classification_problem
from sktime.utils._testing.panel import _make_nested_from_array
from sktime.utils.slope_and_trend import _slope


# Test output format and dimensions.
@pytest.mark.parametrize("n_instances", [1, 3])
@pytest.mark.parametrize("n_timepoints", [10, 20])
@pytest.mark.parametrize("n_intervals", [1, 3, "log", "sqrt", "random"])
@pytest.mark.parametrize(
    "features", [[np.mean], [np.mean, np.median], [np.mean, np.median, np.mean]]
)
def test_output_format_dim(n_instances, n_timepoints, n_intervals, features):
    X = _make_nested_from_array(
        np.ones(n_timepoints), n_instances=n_instances, n_columns=1
    )
    n_rows, n_cols = X.shape
    trans = RandomIntervalFeatureExtractor(n_intervals=n_intervals, features=features)
    Xt = trans.fit_transform(X)
    assert isinstance(Xt, pd.DataFrame)
    assert Xt.shape[0] == n_rows
    assert np.array_equal(Xt.values, np.ones(Xt.shape))


# Check that exception is raised for bad input args.
@pytest.mark.parametrize("bad_n_intervals", [0, "abc", 1.1, -1])
def test_bad_n_intervals(bad_n_intervals):
    X, y = make_classification_problem()
    with pytest.raises(ValueError):
        RandomIntervalFeatureExtractor(n_intervals=bad_n_intervals).fit(X)


@pytest.mark.parametrize(
    "bad_features", [0, "abc", {"a": 1}, (np.median, np.mean), [0, "abc"]]
)
def test_bad_features(bad_features):
    X, y = make_classification_problem()
    with pytest.raises(ValueError):
        RandomIntervalFeatureExtractor(n_intervals=bad_features).fit(X)


# Check specific results
@pytest.mark.parametrize("n_instances", [3, 5])
@pytest.mark.parametrize("n_timepoints", [10, 20])
@pytest.mark.parametrize("n_intervals", [1, 3, "log", "sqrt", "random"])
def test_results(n_instances, n_timepoints, n_intervals):
    X, _ = make_classification_problem(
        n_instances=n_instances, n_timepoints=n_timepoints, return_numpy=True
    )
    transformer = RandomIntervalFeatureExtractor(
        n_intervals=n_intervals, features=[np.mean, np.std]
    )
    Xt = transformer.fit_transform(X)
    Xt = Xt.loc[:, ~Xt.columns.duplicated()]
    # Check results
    intervals = transformer.intervals_
    for start, end in intervals:
        expected_mean = np.mean(X[:, 0, start:end], axis=-1)
        expected_std = np.std(X[:, 0, start:end], axis=-1)

        actual_means = Xt.loc[:, f"{start}_{end}_mean"].to_numpy().ravel()
        actual_stds = Xt.loc[:, f"{start}_{end}_std"].to_numpy().ravel()

        np.testing.assert_array_equal(actual_means, expected_mean)
        np.testing.assert_array_equal(actual_stds, expected_std)


# Test against equivalent pipelines.
def test_different_implementations():
    random_state = 1233
    X_train, y_train = make_classification_problem()

    # Compare with chained transformations.
    tran1 = RandomIntervalSegmenter(n_intervals=1, random_state=random_state)
    tran2 = SeriesToPrimitivesRowTransformer(
        FunctionTransformer(func=np.mean, validate=False), check_transformer=False
    )
    A = tran2.fit_transform(tran1.fit_transform(X_train))

    tran = RandomIntervalFeatureExtractor(
        n_intervals=1, features=[np.mean], random_state=random_state
    )
    B = tran.fit_transform(X_train)

    np.testing.assert_array_almost_equal(A, B)


# Compare with transformer pipeline using TSFeatureUnion.
def test_different_pipelines():
    random_state = 1233
    X_train, y_train = make_classification_problem()
    steps = [
        (
            "segment",
            RandomIntervalSegmenter(n_intervals=1, random_state=random_state),
        ),
        (
            "transform",
            FeatureUnion(
                [
                    (
                        "mean",
                        SeriesToPrimitivesRowTransformer(
                            FunctionTransformer(func=np.mean, validate=False),
                            check_transformer=False,
                        ),
                    ),
                    (
                        "std",
                        SeriesToPrimitivesRowTransformer(
                            FunctionTransformer(func=np.std, validate=False),
                            check_transformer=False,
                        ),
                    ),
                    (
                        "slope",
                        SeriesToPrimitivesRowTransformer(
                            FunctionTransformer(func=_slope, validate=False),
                            check_transformer=False,
                        ),
                    ),
                ]
            ),
        ),
    ]
    pipe = Pipeline(steps)
    a = pipe.fit_transform(X_train)
    tran = RandomIntervalFeatureExtractor(
        n_intervals=1,
        features=[np.mean, np.std, _slope],
        random_state=random_state,
    )
    b = tran.fit_transform(X_train)
    np.testing.assert_array_equal(a, b)
    np.testing.assert_array_equal(pipe.steps[0][1].intervals_, tran.intervals_)



# -*- coding: utf-8 -*-
"""Diverse Representation Canonical Interval Forest Classifier (DrCIF)."""

__author__ = ["Matthew Middlehurst"]
__all__ = ["DrCIF"]

import numpy as np
import math

from joblib import Parallel, delayed
from sklearn import clone
from sklearn.base import BaseEstimator
from sklearn.tree import DecisionTreeClassifier
from sklearn.utils import check_random_state
from sklearn.utils.multiclass import class_distribution


from sktime.contrib._continuous_interval_tree import (
    _drcif_feature,
    ContinuousIntervalTree,
)
from sktime.transformations.panel.catch22 import Catch22
from sktime.utils.validation.panel import check_X, check_X_y
from sktime.classification.base import BaseClassifier


class DrCIF(BaseClassifier):
    """Diverse Representation Canonical Interval Forest Classifier (DrCIF).

    Interval based forest making use of the catch22 feature set on randomly
    selected intervals on the base series, periodogram representation and
    differences representation.

    Overview: Input n series length m
    for each tree
        sample 4 + (sqrt(m)*sqrt(d)) / 3 intervals per representation
        subsample att_subsample_size tsf/catch22 attributes randomly
        randomly select dimension for each interval
        calculate attributes for each interval, concatenate to form new
        data set
        build decision tree on new data set
    ensemble the trees with averaged probability estimates

    This implementation deviates from the original in minor ways. Predictions
    are made using summed probabilities instead of majority vote
    and it does not use the splitting criteria tiny refinement described in
    deng13forest by default.

    Parameters
    ----------
    n_estimators       : int, ensemble size, optional (default to 200)
    n_intervals        : int or size 3 list, number of intervals to extract per
    representation, optional (default to 4 + (sqrt(representation_length)*sqrt(n_dims))
    / 3)
    att_subsample_size : int, number of catch22/tsf attributes to subsample
    per classifier, optional (default to 10)
    min_interval       : int or size 3 list, minimum width of an interval
    per representation, optional (default to 4)
    max_interval       : int or size 3 list, maximum width of an interval
    per representation, optional (default to representation_length / 2)
    n_jobs             : int, optional (default=1)
    The number of jobs to run in parallel for both `fit` and `predict`.
    ``-1`` means using all processors.
    random_state       : int, seed for random, optional (default to no seed)

    Attributes
    ----------
    n_classes      : int, extracted from the data
    n_instances    : int, extracted from the data
    n_dims         : int, extracted from the data
    series_length  : int, extracted from the data
    classifiers    : array of shape = [n_estimators] of DecisionTree
    atts           : array of shape = [n_estimators][att_subsample_size]
    catch22/tsf attribute indexes for all classifiers
    intervals      : array of shape = [n_estimators][n_intervals][2] stores
    indexes of all start and end points for all classifiers
    dims           : array of shape = [n_estimators][n_intervals] stores
    the dimension to extract from for each interval

    Notes
    -----
    For the original Java version, see
    https://github.com/uea-machine-learning/tsml/blob/master/src/main/java
    /tsml/classifiers/interval_based/DrCIF.java

    """

    # Capability tags
    capabilities = {
        "multivariate": True,
        "unequal_length": False,
        "missing_values": False,
        "train_estimate": False,
        "contractable": False,
    }

    def __init__(
        self,
        min_interval=None,
        max_interval=None,
        n_estimators=200,
        n_intervals=None,
        att_subsample_size=10,
        base_estimator=None,
        save_transform_for_estimate=False,
        n_jobs=1,
        random_state=None,
    ):
        self.base_estimator = base_estimator

        self.n_estimators = n_estimators
        self.n_intervals = n_intervals
        self.min_interval = min_interval
        self.max_interval = max_interval
        self.att_subsample_size = att_subsample_size

        self.save_transform_for_estimate = save_transform_for_estimate
        self.random_state = random_state
        self.n_jobs = n_jobs

        # The following set in method fit
        self.n_classes = 0
        self.n_instances = 0
        self.n_dims = 0
        self.series_length = 0
        self.__n_intervals = []
        self.__max_interval = []
        self.__min_interval = []
        self.total_intervals = 0
        self.classifiers = []
        self.atts = []
        self.intervals = []
        self.dims = []
        self.classes_ = []
        self.tree = None

        super(DrCIF, self).__init__()

    def fit(self, X, y):
        """Build a forest of trees from the training set (X, y).

         Uses random intervals and catch22/basic summary features

        Parameters
        ----------
        X : array-like or sparse matrix of shape = [n_instances,n_dimensions,
        series_length] or shape = [n_instances,series_length]
        The training input samples.
        y : array-like, shape =  [n_instances]    The class labels.

        Returns
        -------
        self : object
        """
        X, y = check_X_y(X, y, coerce_to_numpy=True)

        self.n_instances, self.n_dims, self.series_length = X.shape
        self.n_classes = np.unique(y).shape[0]
        self.classes_ = class_distribution(np.asarray(y).reshape(-1, 1))[0][0]

        if self.base_estimator is None or self.base_estimator == "DTC":
            self.tree = DecisionTreeClassifier(criterion="entropy")
        elif self.base_estimator == "CIT":
            self.tree = ContinuousIntervalTree()
        elif isinstance(self.base_estimator, BaseEstimator):
            self.tree = self.base_estimator
        else:
            raise ValueError("DrCIF invalid base estimator given.")

        X_p = np.zeros(
            (
                self.n_instances,
                self.n_dims,
                int(
                    math.pow(2, math.ceil(math.log(self.series_length, 2)))
                    - self.series_length
                ),
            )
        )
        X_p = np.concatenate((X, X_p), axis=2)
        X_p = np.abs(np.fft.fft(X_p)[:, :, : int(X_p.shape[2] / 2)])

        X_d = np.diff(X, 1)

        if self.n_intervals is None:
            self.__n_intervals = [None, None, None]
            self.__n_intervals[0] = 4 + int(
                (math.sqrt(self.series_length) * math.sqrt(self.n_dims)) / 3
            )
            self.__n_intervals[1] = 4 + int(
                (math.sqrt(X_p.shape[2]) * math.sqrt(self.n_dims)) / 3
            )
            self.__n_intervals[2] = 4 + int(
                (math.sqrt(X_d.shape[2]) * math.sqrt(self.n_dims)) / 3
            )
        elif isinstance(self.n_intervals, int):
            self.__n_intervals = [self.n_intervals, self.n_intervals, self.n_intervals]
        elif isinstance(self.n_intervals, list) and len(self.n_intervals) == 3:
            self.__n_intervals = self.n_intervals
        else:
            raise ValueError("DrCIF n_intervals must be an int or list of length 3.")
        for i, n in enumerate(self.__n_intervals):
            if n <= 0:
                self.__n_intervals[i] = 1

        if self.min_interval is None:
            self.__min_interval = [4, 4, 4]
        elif isinstance(self.min_interval, int):
            self.__min_interval = [
                self.min_interval,
                self.min_interval,
                self.min_interval,
            ]
        elif isinstance(self.min_interval, list) and len(self.min_interval) == 3:
            self.__min_interval = self.min_interval
        else:
            raise ValueError("DrCIF min_interval must be an int or list of length 3.")
        if self.series_length < self.__min_interval[0]:
            self.__min_interval[0] = self.series_length
        if X_p.shape[2] < self.__min_interval[1]:
            self.__min_interval[1] = X_p.shape[2]
        if X_d.shape[2] < self.__min_interval[2]:
            self.__min_interval[2] = X_d.shape[2]

        if self.max_interval is None:
            self.__max_interval = [
                self.series_length / 2,
                X_p.shape[2] / 2,
                X_d.shape[2] / 2,
            ]
        elif isinstance(self.max_interval, int):
            self.__max_interval = [
                self.max_interval,
                self.max_interval,
                self.max_interval,
            ]
        elif isinstance(self.max_interval, list) and len(self.max_interval) == 3:
            self.__max_interval = self.max_interval
        else:
            raise ValueError("DrCIF max_interval must be an int or list of length 3.")
        for i, n in enumerate(self.__max_interval):
            if n < self.__min_interval[i]:
                self.__max_interval[i] = self.__min_interval[i]

        self.total_intervals = sum(self.__n_intervals)

        fit = Parallel(n_jobs=self.n_jobs)(
            delayed(self._fit_estimator)(
                X,
                X_p,
                X_d,
                y,
                i,
            )
            for i in range(self.n_estimators)
        )

        self.classifiers, self.intervals, self.dims, self.atts = zip(*fit)

        self._is_fitted = True
        return self

    def predict(self, X):
        """Find predictions for all cases in X. Built on top of predict_proba.

        Parameters
        ----------
        X : The training input samples. array-like or sparse matrix of shape
        = [n_test_instances,n_dimensions,series_length]

        Returns
        -------
        output : array of shape = [n_test_instances]
        """
        rng = check_random_state(self.random_state)
        return np.array(
            [
                self.classes_[int(rng.choice(np.flatnonzero(prob == prob.max())))]
                for prob in self.predict_proba(X)
            ]
        )

    def predict_proba(self, X):
        """Find probability estimates for each class for all cases in X.

        Parameters
        ----------
        X : The training input samples. array-like or sparse matrix of shape
        = [n_test_instances,n_dimensions,series_length]

        Local variables
        ----------
        n_test_instances     : int, number of cases to classify
        series_length    : int, number of attributes in X, must match
        _num_atts determined in fit

        Returns
        -------
        output : array of shape = [n_test_instances, num_classes] of
        probabilities
        """
        self.check_is_fitted()
        X = check_X(X, coerce_to_numpy=True)

        n_test_instances, _, series_length = X.shape
        if series_length != self.series_length:
            raise TypeError(
                "ERROR number of attributes in the train does not match "
                "that in the test data"
            )

        X_p = np.zeros(
            (
                n_test_instances,
                self.n_dims,
                int(
                    math.pow(2, math.ceil(math.log(self.series_length, 2)))
                    - self.series_length
                ),
            )
        )
        X_p = np.concatenate((X, X_p), axis=2)
        X_p = np.abs(np.fft.fft(X_p)[:, :, : int(X_p.shape[2] / 2)])

        X_d = np.diff(X, 1)

        y_probas = Parallel(n_jobs=self.n_jobs)(
            delayed(self._predict_proba_for_estimator)(
                X,
                X_p,
                X_d,
                self.classifiers[i],
                self.intervals[i],
                self.dims[i],
                self.atts[i],
            )
            for i in range(self.n_estimators)
        )

        output = np.sum(y_probas, axis=0) / (
            np.ones(self.n_classes) * self.n_estimators
        )
        return output

    def _fit_estimator(self, X, X_p, X_d, y, idx):
        c22 = Catch22(outlier_norm=True)
        T = [X, X_p, X_d]
        rs = 255 if self.random_state == 0 else self.random_state
        rs = None if self.random_state is None else rs * 37 * (idx + 1)
        rng = check_random_state(rs)

        transformed_x = np.empty(
            shape=(self.att_subsample_size * self.total_intervals, self.n_instances),
            dtype=np.float32,
        )

        atts = rng.choice(29, self.att_subsample_size, replace=False)
        dims = rng.choice(self.n_dims, self.total_intervals, replace=True)
        intervals = np.zeros((self.total_intervals, 2), dtype=int)

        p = 0
        j = 0
        for r in range(0, len(T)):
            transform_length = T[r].shape[2]

            # Find the random intervals for classifier i, transformation r
            # and concatenate features
            for _ in range(0, self.__n_intervals[r]):
                if rng.random() < 0.5:
                    intervals[j][0] = rng.randint(
                        0, transform_length - self.__min_interval[r]
                    )
                    len_range = min(
                        transform_length - intervals[j][0],
                        self.__max_interval[r],
                    )
                    length = (
                        rng.randint(0, len_range - self.__min_interval[r])
                        + self.__min_interval[r]
                    )
                    intervals[j][1] = intervals[j][0] + length
                else:
                    intervals[j][1] = (
                        rng.randint(0, transform_length - self.__min_interval[r])
                        + self.__min_interval[r]
                    )
                    len_range = min(intervals[j][1], self.__max_interval[r])
                    length = (
                        rng.randint(0, len_range - self.__min_interval[r])
                        + self.__min_interval[r]
                        if len_range - self.__min_interval[r] > 0
                        else self.__min_interval[r]
                    )
                    intervals[j][0] = intervals[j][1] - length

                for a in range(0, self.att_subsample_size):
                    transformed_x[p] = _drcif_feature(
                        T[r], intervals[j], dims[j], atts[a], c22
                    )
                    p += 1

                j += 1

        tree = clone(self.tree)
        tree.set_params(random_state=rs)
        transformed_x = transformed_x.T
        transformed_x = transformed_x.round(8)
        transformed_x = np.nan_to_num(transformed_x, False, 0, 0, 0)
        tree.fit(transformed_x, y)

        return [tree, intervals, dims, atts]

    def _predict_proba_for_estimator(
        self, X, X_p, X_d, classifier, intervals, dims, atts
    ):
        c22 = Catch22(outlier_norm=True)
        if isinstance(self.tree, ContinuousIntervalTree):
            return classifier.predict_proba_drcif(
                X, X_p, X_d, c22, self.__n_intervals, intervals, dims, atts
            )
        else:
            T = [X, X_p, X_d]

            transformed_x = np.empty(
                shape=(self.att_subsample_size * self.total_intervals, X.shape[0]),
                dtype=np.float32,
            )

            p = 0
            j = 0
            for r in range(0, len(T)):
                for _ in range(0, self.__n_intervals[r]):
                    for a in range(0, self.att_subsample_size):
                        transformed_x[p] = _drcif_feature(
                            T[r], intervals[j], dims[j], atts[a], c22
                        )
                        p += 1
                    j += 1

            transformed_x = transformed_x.T
            transformed_x.round(8)
            np.nan_to_num(transformed_x, False, 0, 0, 0)

            return classifier.predict_proba(transformed_x)



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
"""Utilities to impute series with missing values."""
__author__ = ["Martin Walter"]
__all__ = ["Imputer"]

from sktime.transformations.base import _SeriesToSeriesTransformer
from sktime.utils.validation.series import check_series
from sktime.forecasting.trend import PolynomialTrendForecaster
from sklearn.utils import check_random_state
from sktime.forecasting.base import ForecastingHorizon
from sklearn.base import clone

import numpy as np
import pandas as pd


class Imputer(_SeriesToSeriesTransformer):
    """Missing value imputation.

    Parameters
    ----------
    method : str, optional (default="drift")
        Method to fill the missing values values.
        * "drift" : drift/trend values by sktime.PolynomialTrendForecaster()
        * "linear" : linear interpolation, by pd.Series.interpolate()
        * "nearest" : use nearest value, by pd.Series.interpolate()
        * "constant" : same constant value (given in arg value) for all NaN
        * "mean" : pd.Series.mean()
        * "median" : pd.Series.median()
        * "backfill"/"bfill" : adapted from pd.Series.fillna()
        * "pad"/"ffill" : adapted from pd.Series.fillna()
        * "random" : random values between pd.Series.min() and .max()
        * "forecaster" : use an sktime Forecaster, given in arg forecaster
    missing_values : int/float/str, optional
        The placeholder for the missing values. All occurrences of
        missing_values will be imputed. Default, None (np.nan)
    value : int/float, optional
        Value to fill NaN, by default None
    forecaster : Any Forecaster based on sktime.BaseForecaster, optinal
        Use a given Forecaster to impute by insample predictions. Before
        fitting, missing data is imputed with method="ffill"/"bfill"
        as heuristic.
    random_state : int/float/str, optional
        Value to set random.seed() if method="random", default None

    Example
    ----------
    >>> from sktime.transformations.series.impute import Imputer
    >>> from sktime.datasets import load_airline
    >>> y = load_airline()
    >>> transformer = Imputer(method="drift")
    >>> y_hat = transformer.fit_transform(y)
    """

    _tags = {
        "fit-in-transform": True,
        "handles-missing-data": True,
        "skip-inverse-transform": True,
    }

    def __init__(
        self,
        method="drift",
        random_state=None,
        value=None,
        forecaster=None,
        missing_values=None,
    ):

        self.method = method
        self.missing_values = missing_values
        self.value = value
        self.forecaster = forecaster
        self.random_state = random_state
        super(Imputer, self).__init__()

    def transform(self, Z, X=None):
        """Transform data.

        Returns a transformed version of Z.

        Parameters
        ----------
        Z : pd.Series, pd.DataFrame

        Returns
        -------
        Z : pd.Series, pd.DataFrame
            Transformed time series(es).
        """
        self.check_is_fitted()
        self._check_method()
        Z = check_series(Z)
        Z = Z.copy()

        # replace missing_values with np.nan
        if self.missing_values:
            Z = Z.replace(to_replace=self.missing_values, value=np.nan)

        if not _has_missing_values(Z):
            return Z

        elif self.method == "random":
            if isinstance(Z, pd.DataFrame):
                for col in Z:
                    Z[col] = Z[col].apply(
                        lambda i: self._get_random(Z[col]) if np.isnan(i) else i
                    )
            else:
                Z = Z.apply(lambda i: self._get_random(Z) if np.isnan(i) else i)
        elif self.method == "constant":
            Z = Z.fillna(value=self.value)
        elif self.method in ["backfill", "bfill", "pad", "ffill"]:
            Z = Z.fillna(method=self.method)
        elif self.method == "drift":
            forecaster = PolynomialTrendForecaster(degree=1)
            Z = _impute_with_forecaster(forecaster, Z)
        elif self.method == "forecaster":
            forecaster = clone(self.forecaster)
            Z = _impute_with_forecaster(forecaster, Z)
        elif self.method == "mean":
            Z = Z.fillna(value=Z.mean())
        elif self.method == "median":
            Z = Z.fillna(value=Z.median())
        elif self.method in ["nearest", "linear"]:
            Z = Z.interpolate(method=self.method)
        else:
            raise ValueError(f"`method`: {self.method} not available.")
        # fill first/last elements of series,
        # as some methods (e.g. "linear") cant impute those
        Z = Z.fillna(method="ffill").fillna(method="backfill")
        return Z

    def _check_method(self):
        if (
            self.value is not None
            and self.method != "constant"
            or self.method == "constant"
            and self.value is None
        ):
            raise ValueError(
                """Imputing with a value can only be
                used if method="constant" and if parameter "value" is not None"""
            )
        elif (
            self.forecaster is not None
            and self.method != "forecaster"
            or self.method == "forecaster"
            and self.forecaster is None
        ):
            raise ValueError(
                """Imputing with a forecaster can only be used if
                method=\"forecaster\" and if arg forecaster is not None"""
            )
        else:
            pass

    def _get_random(self, Z):
        """Create a random int or float value.

        :param Z: Series
        :type Z: pd.Series
        :return: Random int or float between min and max of Z
        :rtype: int/float
        """
        rng = check_random_state(self.random_state)
        # check if series contains only int or int-like values (e.g. 3.0)
        if (Z.dropna() % 1 == 0).all():
            return rng.randint(Z.min(), Z.max())
        else:
            return rng.uniform(Z.min(), Z.max())


def _impute_with_forecaster(forecaster, Z):
    """Use a given forecaster for imputation by in-sample predictions.

    Parameters
    ----------
    forecaster: Forecaster
        Forecaster to use for imputation
    Z : pd.Series or pd.DataFrame
        Series to impute.

    Returns
    -------
    zt : pd.Series or pd.DataFrame
        Series with imputed values.
    """
    if isinstance(Z, pd.Series):
        series = [Z]
    elif isinstance(Z, pd.DataFrame):
        series = [Z[column] for column in Z]

    for z in series:
        # define fh based on index of missing values
        na_index = z.index[z.isna()]
        fh = ForecastingHorizon(values=na_index, is_relative=False)

        # fill NaN before fitting with ffill and backfill (heuristic)
        forecaster.fit(y=z.fillna(method="ffill").fillna(method="backfill"), fh=fh)

        # replace missing values with predicted values
        z[na_index] = forecaster.predict()
    return Z


def _has_missing_values(Z):
    return Z.isnull().to_numpy().any()



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Markus Löning"]
__all__ = ["_HeterogenousEnsembleForecaster"]

from joblib import Parallel
from joblib import delayed
from sklearn.base import clone

from sktime.base import _HeterogenousMetaEstimator
from sktime.forecasting.base._base import DEFAULT_ALPHA
from sktime.forecasting.base._base import BaseForecaster


class _HeterogenousEnsembleForecaster(BaseForecaster, _HeterogenousMetaEstimator):
    """Base class for heterogenous ensemble forecasters"""

    _required_parameters = ["forecasters"]

    def __init__(self, forecasters, n_jobs=None):
        self.forecasters = forecasters
        self.forecasters_ = None
        self.n_jobs = n_jobs
        super(_HeterogenousEnsembleForecaster, self).__init__()

    def _check_forecasters(self):
        if (
            self.forecasters is None
            or len(self.forecasters) == 0
            or not isinstance(self.forecasters, list)
        ):
            raise ValueError(
                "Invalid 'estimators' attribute, 'estimators' should be a list"
                " of (string, estimator) tuples."
            )
        names, forecasters = zip(*self.forecasters)
        # defined by MetaEstimatorMixin
        self._check_names(names)

        has_estimator = any(est not in (None, "drop") for est in forecasters)
        if not has_estimator:
            raise ValueError(
                "All estimators are dropped. At least one is required "
                "to be an estimator."
            )

        for forecaster in forecasters:
            if forecaster not in (None, "drop") and not isinstance(
                forecaster, BaseForecaster
            ):
                raise ValueError(
                    f"The estimator {forecaster.__class__.__name__} should be a "
                    f"Forecaster."
                )
        return names, forecasters

    def _fit_forecasters(self, forecasters, y, X, fh):
        """Fit all forecasters in parallel"""

        def _fit_forecaster(forecaster, y, X, fh):
            """Fit single forecaster"""
            return forecaster.fit(y, X, fh)

        self.forecasters_ = Parallel(n_jobs=self.n_jobs)(
            delayed(_fit_forecaster)(clone(forecaster), y, X, fh)
            for forecaster in forecasters
        )

    def _predict_forecasters(
        self, fh=None, X=None, return_pred_int=False, alpha=DEFAULT_ALPHA
    ):
        """Collect results from forecaster.predict() calls."""
        if return_pred_int:
            raise NotImplementedError()

        return [
            forecaster.predict(fh, X, return_pred_int=return_pred_int, alpha=alpha)
            for forecaster in self.forecasters_
        ]

    def get_params(self, deep=True):
        return self._get_params("forecasters", deep=deep)

    def set_params(self, **params):
        self._set_params("forecasters", **params)
        return self



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Markus Löning"]
__all__ = ["ESTIMATOR_TEST_PARAMS", "EXCLUDE_ESTIMATORS", "EXCLUDED_TESTS"]

import numpy as np

from sktime.registry import (
    ESTIMATOR_TAG_LIST,
    BASE_CLASS_LIST,
    BASE_CLASS_LOOKUP,
    TRANSFORMER_MIXIN_LIST,
)

from pyod.models.knn import KNN
from hcrystalball.wrappers import HoltSmoothingWrapper
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import FunctionTransformer
from sklearn.preprocessing import StandardScaler


from sktime.base import BaseEstimator

from sktime.annotation.adapters import PyODAnnotator
from sktime.classification.compose import ColumnEnsembleClassifier
from sktime.classification.compose import ComposableTimeSeriesForestClassifier
from sktime.classification.dictionary_based import ContractableBOSS
from sktime.classification.dictionary_based import TemporalDictionaryEnsemble
from sktime.classification.hybrid import HIVECOTEV1
from sktime.classification.interval_based import RandomIntervalSpectralForest
from sktime.classification.interval_based._cif import CanonicalIntervalForest
from sktime.classification.interval_based._drcif import DrCIF
from sktime.classification.interval_based import TimeSeriesForestClassifier as TSFC
from sktime.classification.interval_based import SupervisedTimeSeriesForest
from sktime.classification.kernel_based import ROCKETClassifier
from sktime.classification.kernel_based import Arsenal
from sktime.classification.shapelet_based import ShapeletTransformClassifier
from sktime.forecasting.arima import AutoARIMA
from sktime.forecasting.bats import BATS
from sktime.forecasting.compose import DirectTabularRegressionForecaster
from sktime.forecasting.compose import DirRecTimeSeriesRegressionForecaster
from sktime.forecasting.compose import DirectTimeSeriesRegressionForecaster
from sktime.forecasting.compose import DirRecTabularRegressionForecaster
from sktime.forecasting.compose import EnsembleForecaster
from sktime.forecasting.compose import MultioutputTabularRegressionForecaster
from sktime.forecasting.compose import MultioutputTimeSeriesRegressionForecaster
from sktime.forecasting.compose import RecursiveTabularRegressionForecaster
from sktime.forecasting.compose import RecursiveTimeSeriesRegressionForecaster
from sktime.forecasting.compose import StackingForecaster
from sktime.forecasting.compose import TransformedTargetForecaster
from sktime.forecasting.compose import ForecastingPipeline
from sktime.forecasting.compose import MultiplexForecaster
from sktime.forecasting.exp_smoothing import ExponentialSmoothing
from sktime.forecasting.fbprophet import Prophet
from sktime.forecasting.hcrystalball import HCrystalBallForecaster
from sktime.forecasting.model_selection import ForecastingGridSearchCV
from sktime.forecasting.model_selection import ForecastingRandomizedSearchCV
from sktime.forecasting.model_selection import SingleWindowSplitter
from sktime.forecasting.naive import NaiveForecaster
from sktime.forecasting.online_learning import OnlineEnsembleForecaster
from sktime.forecasting.tbats import TBATS
from sktime.forecasting.theta import ThetaForecaster
from sktime.performance_metrics.forecasting import MeanAbsolutePercentageError
from sktime.regression.compose import ComposableTimeSeriesForestRegressor
from sktime.series_as_features.compose import FeatureUnion
from sktime.transformations.panel.compose import ColumnTransformer
from sktime.transformations.panel.compose import (
    SeriesToPrimitivesRowTransformer,
)
from sktime.transformations.panel.compose import SeriesToSeriesRowTransformer
from sktime.transformations.panel.dictionary_based import SFA
from sktime.transformations.panel.interpolate import TSInterpolator
from sktime.transformations.panel.reduce import Tabularizer
from sktime.transformations.panel.shapelets import ContractedShapeletTransform
from sktime.transformations.panel.shapelets import ShapeletTransform
from sktime.transformations.panel.signature_based import SignatureTransformer
from sktime.classification.signature_based import SignatureClassifier
from sktime.transformations.panel.summarize import FittedParamExtractor
from sktime.transformations.panel.tsfresh import TSFreshFeatureExtractor
from sktime.transformations.panel.tsfresh import (
    TSFreshRelevantFeatureExtractor,
)
from sktime.transformations.series.acf import AutoCorrelationTransformer
from sktime.transformations.series.acf import PartialAutoCorrelationTransformer
from sktime.transformations.series.adapt import TabularToSeriesAdaptor
from sktime.transformations.series.detrend import Detrender
from sktime.transformations.series.impute import Imputer
from sktime.transformations.series.compose import OptionalPassthrough
from sktime.transformations.series.outlier_detection import HampelFilter
from sktime.transformations.series.boxcox import BoxCoxTransformer
from sktime.dists_kernels.compose_tab_to_panel import AggrDist
from sktime.dists_kernels.scipy_dist import ScipyDist

# The following estimators currently do not pass all unit tests
# What do they fail? ShapeDTW fails on 3d_numpy_input test, not set up for that
EXCLUDE_ESTIMATORS = [
    "ShapeDTW",
    "ElasticEnsemble",
    "ProximityForest",
    "ProximityStump",
    "ProximityTree",
]


# This is temporary until BaseObject is implemented
DIST_KERNELS_IGNORE_TESTS = [
    "check_fit_updates_state",
    "_make_fit_args",
    "check_fit_returns_self",
    "check_raises_not_fitted_error",
    "check_fit_idempotent",
    "check_fit_does_not_overwrite_hyper_params",
    "check_methods_do_not_change_state",
    "check_persistence_via_pickle",
]


EXCLUDED_TESTS = {
    "ShapeletTransformClassifier": ["check_fit_idempotent"],
    "ContractedShapeletTransform": ["check_fit_idempotent"],
    "HIVECOTEV1": ["check_fit_idempotent", "check_multiprocessing_idempotent"],
    "ScipyDist": DIST_KERNELS_IGNORE_TESTS,
    "AggrDist": DIST_KERNELS_IGNORE_TESTS,
}

# We here configure estimators for basic unit testing, including setting of
# required hyper-parameters and setting of hyper-parameters for faster training.
SERIES_TO_SERIES_TRANSFORMER = StandardScaler()
SERIES_TO_PRIMITIVES_TRANSFORMER = FunctionTransformer(
    np.mean, kw_args={"axis": 0}, check_inverse=False
)
TRANSFORMERS = [
    (
        "transformer1",
        SeriesToSeriesRowTransformer(
            SERIES_TO_SERIES_TRANSFORMER, check_transformer=False
        ),
    ),
    (
        "transformer2",
        SeriesToSeriesRowTransformer(
            SERIES_TO_SERIES_TRANSFORMER, check_transformer=False
        ),
    ),
]
REGRESSOR = LinearRegression()
ANOMALY_DETECTOR = KNN()
TIME_SERIES_CLASSIFIER = TSFC(n_estimators=3)
TIME_SERIES_CLASSIFIERS = [
    ("tsf1", TIME_SERIES_CLASSIFIER),
    ("tsf2", TIME_SERIES_CLASSIFIER),
]
FORECASTER = ExponentialSmoothing()
FORECASTERS = [("ses1", FORECASTER), ("ses2", FORECASTER)]
STEPS_y = [
    ("transformer", Detrender(ThetaForecaster())),
    ("forecaster", NaiveForecaster()),
]
STEPS_X = [
    ("transformer", TabularToSeriesAdaptor(StandardScaler())),
    ("forecaster", NaiveForecaster()),
]
ESTIMATOR_TEST_PARAMS = {
    OnlineEnsembleForecaster: {"forecasters": FORECASTERS},
    FeatureUnion: {"transformer_list": TRANSFORMERS},
    DirectTabularRegressionForecaster: {"estimator": REGRESSOR},
    MultioutputTabularRegressionForecaster: {"estimator": REGRESSOR},
    RecursiveTabularRegressionForecaster: {"estimator": REGRESSOR},
    DirRecTabularRegressionForecaster: {"estimator": REGRESSOR},
    DirectTimeSeriesRegressionForecaster: {
        "estimator": make_pipeline(Tabularizer(), REGRESSOR)
    },
    RecursiveTimeSeriesRegressionForecaster: {
        "estimator": make_pipeline(Tabularizer(), REGRESSOR)
    },
    MultioutputTimeSeriesRegressionForecaster: {
        "estimator": make_pipeline(Tabularizer(), REGRESSOR)
    },
    DirRecTimeSeriesRegressionForecaster: {
        "estimator": make_pipeline(Tabularizer(), REGRESSOR)
    },
    TransformedTargetForecaster: {"steps": STEPS_y},
    ForecastingPipeline: {"steps": STEPS_X},
    EnsembleForecaster: {"forecasters": FORECASTERS},
    StackingForecaster: {"forecasters": FORECASTERS, "final_regressor": REGRESSOR},
    Detrender: {"forecaster": FORECASTER},
    ForecastingGridSearchCV: {
        "forecaster": NaiveForecaster(strategy="mean"),
        "cv": SingleWindowSplitter(fh=1),
        "param_grid": {"window_length": [2, 5]},
        "scoring": MeanAbsolutePercentageError(symmetric=True),
    },
    ForecastingRandomizedSearchCV: {
        "forecaster": NaiveForecaster(strategy="mean"),
        "cv": SingleWindowSplitter(fh=1),
        "param_distributions": {"window_length": [2, 5]},
        "scoring": MeanAbsolutePercentageError(symmetric=True),
    },
    TabularToSeriesAdaptor: {"transformer": StandardScaler()},
    ColumnEnsembleClassifier: {
        "estimators": [
            (name, estimator, 0) for (name, estimator) in TIME_SERIES_CLASSIFIERS
        ]
    },
    FittedParamExtractor: {
        "forecaster": FORECASTER,
        "param_names": ["initial_level"],
    },
    SeriesToPrimitivesRowTransformer: {
        "transformer": SERIES_TO_PRIMITIVES_TRANSFORMER,
        "check_transformer": False,
    },
    SeriesToSeriesRowTransformer: {
        "transformer": SERIES_TO_SERIES_TRANSFORMER,
        "check_transformer": False,
    },
    ColumnTransformer: {
        "transformers": [(name, estimator, [0]) for name, estimator in TRANSFORMERS]
    },
    AutoARIMA: {
        "d": 0,
        "suppress_warnings": True,
        "max_p": 2,
        "max_q": 2,
        "seasonal": False,
    },
    MultiplexForecaster: {
        "forecasters": [
            ("Naive_mean", NaiveForecaster(strategy="mean")),
            ("Naive_last", NaiveForecaster(strategy="last")),
            ("Naive_drift", NaiveForecaster(strategy="drift")),
        ],
        "selected_forecaster": "Naive_mean",
    },
    ShapeletTransformClassifier: {
        "n_estimators": 3,
        "transform_contract_in_mins": 0.075,
    },
    ContractedShapeletTransform: {"time_contract_in_mins": 0.075},
    ShapeletTransform: {
        "max_shapelets_to_store_per_class": 1,
        "min_shapelet_length": 3,
        "max_shapelet_length": 4,
    },
    SignatureTransformer: {
        "augmentation_list": ("basepoint", "addtime"),
        "depth": 3,
        "window_name": "global",
    },
    SignatureClassifier: {
        "augmentation_list": ("basepoint", "addtime"),
        "depth": 3,
        "window_name": "global",
    },
    ROCKETClassifier: {"num_kernels": 100},
    Arsenal: {"num_kernels": 100},
    HIVECOTEV1: {
        "stc_params": {"n_estimators": 2, "transform_contract_in_mins": 0.025},
        "tsf_params": {"n_estimators": 2},
        "rise_params": {"n_estimators": 2},
        "cboss_params": {"n_parameter_samples": 6, "max_ensemble_size": 2},
    },
    TSFreshFeatureExtractor: {"disable_progressbar": True, "show_warnings": False},
    TSFreshRelevantFeatureExtractor: {
        "disable_progressbar": True,
        "show_warnings": False,
        "fdr_level": 0.01,
    },
    TSInterpolator: {"length": 10},
    RandomIntervalSpectralForest: {"n_estimators": 3, "acf_lag": 10, "min_interval": 5},
    SFA: {"return_pandas_data_series": True},
    ContractableBOSS: {"n_parameter_samples": 25, "max_ensemble_size": 5},
    TemporalDictionaryEnsemble: {
        "n_parameter_samples": 25,
        "max_ensemble_size": 5,
        "randomly_selected_params": 20,
    },
    TSFC: {"n_estimators": 3},
    ComposableTimeSeriesForestClassifier: {"n_estimators": 3},
    ComposableTimeSeriesForestRegressor: {"n_estimators": 3},
    SupervisedTimeSeriesForest: {"n_estimators": 3},
    CanonicalIntervalForest: {"n_estimators": 3},
    DrCIF: {"n_estimators": 3},
    HCrystalBallForecaster: {"model": HoltSmoothingWrapper()},
    BATS: {
        "use_box_cox": False,
        "use_trend": False,
        "use_damped_trend": False,
        "sp": [],
        "use_arma_errors": False,
        "n_jobs": 1,
    },
    TBATS: {
        "use_box_cox": False,
        "use_trend": False,
        "use_damped_trend": False,
        "sp": [],
        "use_arma_errors": False,
        "n_jobs": 1,
    },
    Prophet: {
        "n_changepoints": 0,
        "yearly_seasonality": False,
        "weekly_seasonality": False,
        "daily_seasonality": False,
        "uncertainty_samples": 1000,
        "verbose": False,
    },
    PartialAutoCorrelationTransformer: {"n_lags": 1},
    AutoCorrelationTransformer: {"n_lags": 1},
    Imputer: {"method": "mean"},
    HampelFilter: {"window_length": 3},
    OptionalPassthrough: {"transformer": BoxCoxTransformer(), "passthrough": True},
    AggrDist: {"transformer": ScipyDist()},
    PyODAnnotator: {"estimator": ANOMALY_DETECTOR},
}

# We use estimator tags in addition to class hierarchies to further distinguish
# estimators into different categories. This is useful for defining and running
# common tests for estimators with the same tags.
VALID_ESTIMATOR_TAGS = tuple(ESTIMATOR_TAG_LIST)

# These methods should not change the state of the estimator, that is, they should
# not change fitted parameters or hyper-parameters. They are also the methods that
# "apply" the fitted estimator to data and useful for checking results.
NON_STATE_CHANGING_METHODS = (
    "predict",
    "predict_proba",
    "decision_function",
    "transform",
    "inverse_transform",
)

# The following gives a list of valid estimator base classes.
VALID_TRANSFORMER_TYPES = tuple(TRANSFORMER_MIXIN_LIST)

VALID_ESTIMATOR_BASE_TYPES = tuple(BASE_CLASS_LIST)

VALID_ESTIMATOR_TYPES = (
    BaseEstimator,
    *VALID_ESTIMATOR_BASE_TYPES,
    *VALID_TRANSFORMER_TYPES,
)

VALID_ESTIMATOR_BASE_TYPE_LOOKUP = BASE_CLASS_LOOKUP



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-

__author__ = ["Markus Löning"]
__all__ = []

import pytest

from sktime.tests._config import EXCLUDE_ESTIMATORS
from sktime.tests._config import VALID_ESTIMATOR_BASE_TYPE_LOOKUP
from sktime.tests._config import VALID_ESTIMATOR_TYPES
from sktime.utils import all_estimators


@pytest.mark.parametrize("return_names", [True, False])
def test_all_estimators_return_names(return_names):
    estimators = all_estimators(return_names=return_names)
    assert isinstance(estimators, list)
    assert len(estimators) > 0

    if return_names:
        assert all([isinstance(estimator, tuple) for estimator in estimators])
        names, estimators = list(zip(*estimators))
        assert all([isinstance(name, str) for name in names])
        assert all(
            [name == estimator.__name__ for name, estimator in zip(names, estimators)]
        )

    assert all([isinstance(estimator, type) for estimator in estimators])


@pytest.mark.parametrize("exclude_estimators", ["NaiveForecaster", EXCLUDE_ESTIMATORS])
def test_all_estimators_exclude_estimators(exclude_estimators):
    estimators = all_estimators(
        return_names=True, exclude_estimators=exclude_estimators
    )
    assert isinstance(estimators, list)
    assert len(estimators) > 0
    names, estimators = list(zip(*estimators))

    if not isinstance(exclude_estimators, list):
        exclude_estimators = [exclude_estimators]
    for estimator in exclude_estimators:
        assert estimator not in names


@pytest.mark.parametrize(
    "estimator_type", [*VALID_ESTIMATOR_BASE_TYPE_LOOKUP.keys(), *VALID_ESTIMATOR_TYPES]
)
def test_all_estimators_filter_estimator_types(estimator_type):
    estimators = all_estimators(estimator_types=estimator_type, return_names=False)
    assert isinstance(estimators, list)
    # assert len(estimators) > 0

    if isinstance(estimator_type, str):
        estimator_type = VALID_ESTIMATOR_BASE_TYPE_LOOKUP[estimator_type]

    assert all([issubclass(estimator, estimator_type) for estimator in estimators])



# -*- coding: utf-8 -*-
import numpy as np
import os
from sklearn.metrics import accuracy_score
from sktime.contrib.classification_experiments import write_results_to_uea_format


class ElasticEnsemblePostProcess:
    __author__ = "Jason Lines"

    def __init__(
        self,
        results_path,
        dataset_name,
        distance_measures="all",
        resample_id=0,
        alpha=1,
    ):
        """

        Args:
            results_path: String - path to folder storing the results to be read into the ensemble
            dataset_name: String - the name of the dataset that this ensemble will post process results for
            distance_measures: 'all' or list of Strings - default 'all'. 'all' sets classifier to use all default constituent classifiers,
                                else a list is provided of classifiers to include. Note these names must match the names of
                                the subdirs in the folder located at results_parh
            resample_id: default = 0 - to identify the deterministic seed used for resampling experiments.
                         A resampled_id of 0 demonstrates default train/test split were used to create results
            alpha: float/double - default=1.0. Used to exponentiate the confidence of constituent classifiers when making test predictions
        """
        self.results_path = results_path
        self.dataset_name = dataset_name
        self.resample_id = resample_id
        if distance_measures == "all":
            self.distance_measures = [
                "dtw",
                "ddtw",
                "wdtw",
                "wddtw",
                "lcss",
                "erp",
                "msm",
            ]
        else:
            self.distance_measures = distance_measures
        self.alpha = alpha

        # load in train information
        self.train_accs_by_classifier = np.zeros(len(self.distance_measures))
        self.train_dists_by_classifier = []
        self.test_dists_by_classifier = []
        self.ee_train_dists = None
        self.ee_test_dists = None

        self.actual_train_class_vals = None
        self.actual_test_class_vals = None
        self.classes_ = None

        num_classes = None
        num_ins = None
        class_vals = None
        # load train

        for c_id in range(len(self.distance_measures)):
            file_path = (
                self.results_path
                + self.distance_measures[c_id]
                + "/Predictions/"
                + self.dataset_name
                + "/trainFold"
                + str(self.resample_id)
                + ".csv"
            )
            with open(file_path, "r") as f:
                lines = f.readlines()
                third_line = lines[2].split(",")
                self.train_accs_by_classifier[c_id] = float(third_line[0].strip())
                this_class_vals = (
                    third_line[-1].strip().replace("[", "").replace("]", "").split(" ")
                )
                this_num_classes = len(this_class_vals)
                this_num_ins = len(lines) - 3
                if class_vals is None:
                    class_vals = this_class_vals
                    self.classes_ = np.array(this_class_vals)
                    num_classes = this_num_classes
                    num_ins = this_num_ins

                elif this_class_vals != class_vals:
                    raise ValueError(
                        "Class value mismatch when loading train file for "
                        + str(self.distance_measures[c_id])
                        + " and "
                        + self.dataset_name
                    )
                elif this_num_ins != num_ins:
                    raise ValueError(
                        "Inconsistent number of predictions in constituent training files: first spotted "
                        "in train file for "
                        + str(self.distance_measures[c_id])
                        + " and "
                        + self.dataset_name
                    )

                this_dists = np.empty((num_ins, num_classes))
                this_actual_train_class_vals = []
                for i in range(num_ins):
                    split_line = lines[i + 3].strip().split(",")
                    this_actual_train_class_vals.append(split_line[0].strip())
                    for c in range(num_classes):
                        this_dists[i][c] = (
                            np.power(float(split_line[c + 3]), self.alpha)
                            * self.train_accs_by_classifier[c_id]
                        )
                if self.actual_train_class_vals is None:
                    self.actual_train_class_vals = this_actual_train_class_vals
                elif self.actual_train_class_vals != this_actual_train_class_vals:
                    raise ValueError(
                        "Class values in files no not match for train - first spotted for "
                        + str(self.distance_measures[c_id])
                    )

            if self.ee_train_dists is None:
                self.ee_train_dists = this_dists
            else:
                self.ee_train_dists = np.add(self.ee_train_dists, this_dists)
            self.train_dists_by_classifier.append(this_dists)
        self.ee_train_dists = np.divide(
            self.ee_train_dists, sum(self.train_accs_by_classifier)
        )

        # load test
        num_test_ins = None
        for c_id in range(len(self.distance_measures)):
            file_path = (
                self.results_path
                + self.distance_measures[c_id]
                + "/Predictions/"
                + self.dataset_name
                + "/testFold"
                + str(self.resample_id)
                + ".csv"
            )
            with open(file_path, "r") as f:
                lines = f.readlines()
                third_line = lines[2].split(",")
                this_class_vals = (
                    third_line[-1].strip().replace("[", "").replace("]", "").split(" ")
                )
                this_num_ins = len(lines) - 3
                if this_class_vals != class_vals:
                    raise ValueError(
                        "Class value mismatch when loading test file for "
                        + str(self.distance_measures[c_id])
                        + " and "
                        + self.dataset_name
                    )
                if num_test_ins is None:
                    num_test_ins = this_num_ins
                elif num_test_ins != this_num_ins:
                    raise ValueError(
                        "Inconsistent number of predictions in constituent test files: first spotted "
                        "in train file for "
                        + str(self.distance_measures[c_id])
                        + " and "
                        + self.dataset_name
                    )
                this_dists = np.empty((num_ins, num_classes))
                this_actual_test_class_vals = []
                for i in range(num_ins):
                    split_line = lines[i + 3].strip().split(",")
                    this_actual_test_class_vals.append(split_line[0].strip())
                    for c in range(num_classes):
                        this_dists[i][c] = (
                            np.power(float(split_line[c + 3]), self.alpha)
                            * self.train_accs_by_classifier[c_id]
                        )
                if self.actual_test_class_vals is None:
                    self.actual_test_class_vals = this_actual_test_class_vals
                elif self.actual_test_class_vals != this_actual_test_class_vals:
                    raise ValueError(
                        "Class values in files no not match for test - first spotted for "
                        + str(self.distance_measures[c_id])
                    )

            if self.ee_test_dists is None:
                self.ee_test_dists = this_dists
            else:
                self.ee_test_dists = np.add(self.ee_test_dists, this_dists)
            self.test_dists_by_classifier.append(this_dists)

        self.ee_test_dists = np.divide(
            self.ee_test_dists, sum(self.train_accs_by_classifier)
        )

    def write_files(
        self,
        output_results_path,
        output_classifier_name="EE",
        write_train=True,
        write_test=True,
        overwrite=False,
    ):
        """

        Args:
            output_results_path: String - path to where output results will be written
            output_classifier_name: String - the name of the composite ensemble classifier in the output files
            write_train: boolean - true will write train files for the ensemble, false will skip training files
            write_test: boolean - true will write test files for the ensemble, false will skip test files
            overwrite: boolean - if true, any existing train/test files will be over-written. False prevents file overwriting


        """
        if write_train is False and write_test is False:
            print(
                "Train and test writing both set to false - method will terminate without doing anything"
            )
            return

        if not overwrite:
            if write_train:
                full_path = (
                    str(output_results_path)
                    + "/"
                    + str(output_classifier_name)
                    + "/Predictions/"
                    + str(self.dataset_name)
                    + "/trainFold"
                    + str(self.resample_id)
                    + ".csv"
                )
                if os.path.exists(full_path):
                    print(
                        full_path
                        + " already exists and overwrite set to false, not writing Train",
                        Warning,
                    )
                    write_train = False

            if write_test is True:
                full_path = (
                    str(output_results_path)
                    + "/"
                    + str(output_classifier_name)
                    + "/Predictions/"
                    + str(self.dataset_name)
                    + "/testFold"
                    + str(self.resample_id)
                    + ".csv"
                )
                if os.path.exists(full_path):
                    print(
                        full_path
                        + " already exists and overwrite set to false, not writing Test"
                    )
                    write_test = False

        if write_train is False and write_test is False:
            print(
                "Train and test files both already exist and overwrite set to false - method will terminate without doing anything"
            )
            return

        """
        file_format = None
        if os.path.exists(problem_path + self.dataset_name + '/' + self.dataset_name+ '_TRAIN.ts') and os.path.exists(problem_path + self.dataset_name + '/' + self.dataset_name+ '_TEST.ts'):
            train_x, train_y = loader.load_from_tsfile_to_dataframe(problem_path + self.dataset_name + '/' + self.dataset_name + '_TRAIN.ts')
            test_x, test_y = loader.load_from_tsfile_to_dataframe(problem_path + self.dataset_name + '/' + self.dataset_name + '_TEST.ts')
        elif os.path.exists(problem_path + self.dataset_name + '/' + self.dataset_name+ '_TRAIN.arff') and os.path.exists(problem_path + self.dataset_name + '/' + self.dataset_name+ '_TEST.arff'):
            train_x, train_y = loader.load_from_arff_to_dataframe(problem_path + self.dataset_name + '/' + self.dataset_name + '_TRAIN.ts')
            test_x, test_y = loader.load_from_arff_to_dataframe(problem_path + self.dataset_name + '/' + self.dataset_name + '_TEST.ts')
        elif os.path.exists(problem_path + self.dataset_name + '/' + self.dataset_name + '_TRAIN.tsv') and os.path.exists(problem_path + self.dataset_name + '/' + self.dataset_name + '_TEST.tsv'):
            train_x, train_y = loader.load_from_ucr_tsv_to_dataframe(problem_path + self.dataset_name + '/' + self.dataset_name + '_TRAIN.ts')
            test_x, test_y = loader.load_from_ucr_tsv_to_dataframe(problem_path + self.dataset_name + '/' + self.dataset_name + '_TEST.ts')
        else:
            raise ValueError("No dataset found for "+self.dataset_name)
        """
        if write_train:
            train_probs = self.ee_train_dists
            train_preds = self.classes_[np.argmax(train_probs, axis=1)]
            acc = accuracy_score(self.actual_train_class_vals, train_preds)
            second = str(self.distance_measures)
            third = (
                str(acc)
                + ",NA,NA,-1,-1,"
                + str(len(self.classes_))
                + ","
                + str(self.classes_)
            )
            write_results_to_uea_format(
                second_line=second,
                third_line=third,
                output_path=output_results_path,
                classifier_name=output_classifier_name,
                resample_seed=self.resample_id,
                predicted_class_vals=train_preds,
                actual_probas=train_probs,
                dataset_name=self.dataset_name,
                actual_class_vals=self.actual_train_class_vals,
                split="TRAIN",
            )

        if write_test:
            test_probs = self.ee_test_dists
            test_preds = self.classes_[np.argmax(test_probs, axis=1)]
            acc = accuracy_score(self.actual_test_class_vals, test_preds)
            second = str(self.distance_measures)
            third = (
                str(acc)
                + ",NA,NA,-1,-1,"
                + str(len(self.classes_))
                + ","
                + str(self.classes_)
            )
            write_results_to_uea_format(
                second_line=second,
                third_line=third,
                output_path=output_results_path,
                classifier_name=output_classifier_name,
                resample_seed=self.resample_id,
                predicted_class_vals=test_preds,
                actual_probas=test_probs,
                dataset_name=self.dataset_name,
                actual_class_vals=self.actual_test_class_vals,
                split="TEST",
            )



# -*- coding: utf-8 -*-
import numpy as np

from sktime.clustering.partitioning._averaging_metrics import (
    BarycenterAveraging,
    MeanAveraging,
)
from sktime.clustering.tests._clustering_tests import generate_univaritate_series


def test_barycenter_averaging():
    rng = np.random.RandomState(0)
    X = generate_univaritate_series(n=100, size=5, rng=rng, dtype=np.int)

    BCA = BarycenterAveraging(X)
    BCA.average()


def test_mean_averaging():
    rng = np.random.RandomState(1)
    X = generate_univaritate_series(n=100, size=5, rng=rng, dtype=np.int)

    mean = MeanAveraging(X)
    average = mean.average()
    assert np.array_equal(np.array([-0.02, 0.01, 0.05, 0.11, 0.03]), average)



# -*- coding: utf-8 -*-
__all__ = [
    "ShapeletTransformClassifier",
    "MrSEQLClassifier",
]

from sktime.classification.shapelet_based._stc import ShapeletTransformClassifier
from sktime.classification.shapelet_based.mrseql.mrseql import MrSEQLClassifier



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["Tomasz Chodakowski", "Ryan Kuhns"]

import pytest
import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype
from sktime.utils._testing.series import _make_series
from sktime.performance_metrics.forecasting import (
    MeanAbsoluteScaledError,
    MedianAbsoluteScaledError,
    MeanSquaredScaledError,
    MedianSquaredScaledError,
    MeanAbsoluteError,
    MeanSquaredError,
    MedianAbsoluteError,
    MedianSquaredError,
    MeanAbsolutePercentageError,
    MedianAbsolutePercentageError,
    MeanSquaredPercentageError,
    MedianSquaredPercentageError,
    MeanRelativeAbsoluteError,
    MedianRelativeAbsoluteError,
    GeometricMeanRelativeAbsoluteError,
    GeometricMeanRelativeSquaredError,
    MeanAsymmetricError,
    RelativeLoss,
    mean_absolute_scaled_error,
    median_absolute_scaled_error,
    mean_squared_scaled_error,
    median_squared_scaled_error,
    mean_absolute_error,
    mean_squared_error,
    median_absolute_error,
    median_squared_error,
    mean_absolute_percentage_error,
    median_absolute_percentage_error,
    mean_squared_percentage_error,
    median_squared_percentage_error,
    mean_relative_absolute_error,
    median_relative_absolute_error,
    geometric_mean_relative_absolute_error,
    geometric_mean_relative_squared_error,
    mean_asymmetric_error,
    relative_loss,
)
from sktime.performance_metrics.tests._config import RANDOM_SEED

# For multiple comparisons of equality between functions and classes
rng = np.random.default_rng(RANDOM_SEED)
RANDOM_STATES = rng.integers(0, 1000000, size=5).tolist()

# Create specific test series to verify calculated performance metrics match
# those calculated externally
Y1 = np.array(
    [
        0.626832772836215,
        0.783382993377663,
        0.745780385700732,
        1.06737808331213,
        1.69664933579028,
        2.08627141338732,
        1.78023192557434,
        1.58568920200064,
        2.08902410668301,
        2.51472070324453,
        2.47425419784015,
        2.27275916300358,
        1.92803852608368,
        1.64662766528414,
        1.7028471682496,
        1.62051042240568,
        2.03642032341352,
        2.36019377457168,
        2.39730479510699,
        2.69699728045652,
        2.41172828049954,
        2.37679353181132,
        1.99603448413176,
        2.53946033171028,
        2.16285521091308,
        1.70889477546947,
        1.52488156869114,
        1.8369477471545,
        1.8225935878131,
        1.64685504990138,
        1.36106553603259,
        1.20252674753628,
        1.33235953453508,
        1.70560866839458,
        2.25722026784685,
        1.84446872239422,
    ]
)

Y2 = pd.Series(
    [
        0.982136629140069,
        1.45950325745833,
        1.42708285946536,
        2.10474124388042,
        2.12958738712948,
        1.94254184770726,
        2.24111458763484,
        2.68784805815518,
        2.97248086366361,
        3.27426914233203,
        3.16674535150384,
        2.933698752984,
        3.18393847027259,
        3.43030921792323,
        3.21901076902567,
        2.51266154720592,
        2.52702260323378,
        2.4241798970835,
        1.91495784087606,
        1.49993972682056,
        1.66460722130508,
        1.72380847201769,
        1.45265679700175,
        1.54961689438936,
        1.40262473301413,
        1.50833698230433,
        1.17807171492728,
        1.37642259034361,
        1.19122274092639,
        1.72766650406602,
        2.01019283258555,
        1.70144149287405,
        1.40552850108184,
        1.22336047820607,
        1.58882703694742,
        1.68674857175401,
    ]
)
# Data for this test case borrower from Rob Hyndman's excel workbook
# demonstrating how to calculate MASE
Y3 = np.array(
    [
        0,
        2,
        0,
        1,
        0,
        11,
        0,
        0,
        0,
        0,
        2,
        0,
        6,
        3,
        0,
        0,
        0,
        0,
        0,
        7,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        3,
        1,
        0,
        0,
        1,
        0,
        1,
        0,
        0,
    ]
)
Y1_TRAIN, Y1_TEST = Y1[:24], Y1[24:]
Y2_TRAIN, Y2_TEST = Y2[:24], Y2[24:]
Y3_TRAIN, Y3_TEST = Y3[:24], Y3[24:]

Y_TEST_CASES = {
    "test_case_1": {"train": Y1_TRAIN, "test": Y1_TEST},
    "test_case_2": {"train": Y2_TRAIN, "test": Y2_TEST},
    "test_case_3": {"train": Y3_TRAIN, "test": Y3_TEST},
    # Multivariate test case
    "test_case_4": {
        "train": np.vstack([Y1_TRAIN, Y2_TRAIN]),
        "test": np.vstack([Y1_TEST, Y2_TEST]),
    },
}

# Dictionary mapping functions to the true loss values to verify the sktime
# metrics are performing as expected. True loss values were calculated
# manually outside of sktime in Excel.
LOSS_RESULTS = {
    "mean_absolute_scaled_error": {
        "test_case_1": 1.044427857,
        "test_case_2": 0.950832524,
        "test_case_3": 0.33045977,
        "func": mean_absolute_scaled_error,
        "class": MeanAbsoluteScaledError(),
    },
    "median_absolute_scaled_error": {
        "test_case_1": 0.997448587,
        "test_case_2": 0.975921875,
        "test_case_3": 1.0,
        "func": median_absolute_scaled_error,
        "class": MedianAbsoluteScaledError(),
    },
    "root_mean_squared_scaled_error": {
        "test_case_1": 1.001351033,
        "test_case_2": 0.854561506,
        "test_case_3": 0.289374954,
        "func": mean_squared_scaled_error,
        "class": MeanSquaredScaledError(square_root=True),
    },
    "root_median_squared_scaled_error": {
        "test_case_1": 0.998411526,
        "test_case_2": 0.990760662,
        "test_case_3": 1.0,
        "func": median_squared_scaled_error,
        "class": MedianSquaredScaledError(square_root=True),
    },
    "mean_absolute_error": {
        "test_case_1": 0.285709251,
        "test_case_2": 0.252975912,
        "test_case_3": 0.833333333,
        "func": mean_absolute_error,
        "class": MeanAbsoluteError(),
    },
    "mean_squared_error": {
        "test_case_1": 0.103989049,
        "test_case_2": 0.07852696,
        "test_case_3": 1.5,
        "func": mean_squared_error,
        "class": MeanSquaredError(),
    },
    "root_mean_squared_error": {
        "test_case_1": 0.322473331,
        "test_case_2": 0.280226623,
        "test_case_3": 1.224744871,
        "func": mean_squared_error,
        "class": MeanSquaredError(square_root=True),
    },
    "median_absolute_error": {
        "test_case_1": 0.298927846,
        "test_case_2": 0.240438602,
        "test_case_3": 1.0,
        "func": median_absolute_error,
        "class": MedianAbsoluteError(),
    },
    "median_squared_error": {
        "test_case_1": 0.089530473,
        "test_case_2": 0.059582098,
        "test_case_3": 1.0,
        "func": median_squared_error,
        "class": MedianSquaredError(),
    },
    "root_median_squared_error": {
        "test_case_1": 0.299216432,
        "test_case_2": 0.244094445,
        "test_case_3": 1.0,
        "func": median_squared_error,
        "class": MedianSquaredError(square_root=True),
    },
    "symmetric_mean_absolute_percentage_error": {
        "test_case_1": 0.16206745335345693,
        "test_case_2": 0.17096048184064724,
        "test_case_3": 1.0833333333333333,
        "func": mean_absolute_percentage_error,
        "class": MeanAbsolutePercentageError(symmetric=True),
    },
    "symmetric_median_absolute_percentage_error": {
        "test_case_1": 0.17291559217102262,
        "test_case_2": 0.15323286657516913,
        "test_case_3": 1.5,
        "func": median_absolute_percentage_error,
        "class": MedianAbsolutePercentageError(symmetric=True),
    },
    "mean_absolute_percentage_error": {
        "test_case_1": 0.16426360194846226,
        "test_case_2": 0.16956968442429066,
        "test_case_3": 1125899906842624.2,
        "func": mean_absolute_percentage_error,
        "class": MeanAbsolutePercentageError(symmetric=False),
    },
    "median_absolute_percentage_error": {
        "test_case_1": 0.17200352348889714,
        "test_case_2": 0.1521891319356885,
        "test_case_3": 1.0,
        "func": median_absolute_percentage_error,
        "class": MedianAbsolutePercentageError(symmetric=False),
    },
    "mean_squared_percentage_error": {
        "test_case_1": 0.03203423036447087,
        "test_case_2": 0.03427486821803671,
        "test_case_3": 5.070602400912918e30,
        "func": mean_squared_percentage_error,
        "class": MeanSquaredPercentageError(symmetric=False),
    },
    "median_squared_percentage_error": {
        "test_case_1": 0.029589708748632582,
        "test_case_2": 0.023172298452886965,
        "test_case_3": 1.0,
        "func": median_squared_percentage_error,
        "class": MedianSquaredPercentageError(symmetric=False),
    },
    "root_mean_squared_percentage_error": {
        "test_case_1": 0.17898108940463758,
        "test_case_2": 0.18513472990780716,
        "test_case_3": 2251799813685248.0,
        "func": mean_squared_percentage_error,
        "class": MeanSquaredPercentageError(square_root=True, symmetric=False),
    },
    "root_median_squared_percentage_error": {
        "test_case_1": 0.17201659439900727,
        "test_case_2": 0.15222450017289255,
        "test_case_3": 1.0,
        "func": median_squared_percentage_error,
        "class": MedianSquaredPercentageError(square_root=True, symmetric=False),
    },
    "mean_relative_absolute_error": {
        "test_case_1": 0.485695805,
        "test_case_2": 0.477896036,
        "test_case_3": 0.875,
        "func": mean_relative_absolute_error,
        "class": MeanRelativeAbsoluteError(),
    },
    "median_relative_absolute_error": {
        "test_case_1": 0.411364556,
        "test_case_2": 0.453437859,
        "test_case_3": 1.0,
        "func": median_relative_absolute_error,
        "class": MedianRelativeAbsoluteError(),
    },
    "geometric_mean_relative_absolute_error": {
        "test_case_1": 0.363521894,
        "test_case_2": 0.402438951,
        "test_case_3": 3.6839e-07,
        "func": geometric_mean_relative_absolute_error,
        "class": GeometricMeanRelativeAbsoluteError(),
    },
    "geometric_mean_relative_squared_error": {
        "test_case_1": 0.132148167,
        "test_case_2": 0.161957109,
        "test_case_3": 4.517843023201426e-07,
        "func": geometric_mean_relative_squared_error,
        "class": GeometricMeanRelativeSquaredError(),
    },
    "mean_aymmetric_error": {
        "test_case_1": 0.17139968,
        "test_case_2": 0.163956601,
        "test_case_3": 1.000000,
        "func": mean_asymmetric_error,
        "class": MeanAsymmetricError(),
    },
    "relative_loss": {
        "test_case_1": 0.442644622,
        "test_case_2": 0.416852592,
        "test_case_3": 1.315789474,
        "func": relative_loss,
        "class": RelativeLoss(),
    },
}


def _call_metrics(metric_func, metric_class, y_true, y_pred, y_train, y_pred_benchmark):
    """Call function and class metrics and return results"""
    class_attrs = metric_class.get_params()
    function_metric = metric_func(
        y_true,
        y_pred,
        y_train=y_train,
        y_pred_benchmark=y_pred_benchmark,
        **class_attrs,
    )
    class_metric = metric_class(
        y_true,
        y_pred,
        y_train=y_train,
        y_pred_benchmark=y_pred_benchmark,
    )
    return function_metric, class_metric


@pytest.mark.parametrize("metric_func_name", LOSS_RESULTS.keys())
@pytest.mark.parametrize("n_test_case", [1, 2, 3])
def test_univariate_loss_expected_zero(n_test_case, metric_func_name):
    # Test cases where the expected loss is zero for perfect forecast.

    metric_class = LOSS_RESULTS[metric_func_name]["class"]
    metric_func = LOSS_RESULTS[metric_func_name]["func"]

    y_true = Y_TEST_CASES[f"test_case_{n_test_case}"]["test"]
    y_train = Y_TEST_CASES[f"test_case_{n_test_case}"]["train"]

    # Setting test case of perfect forecast and benchmark
    true_loss = 0
    y_pred = y_true
    y_pred_benchmark = y_true

    function_loss, class_loss = _call_metrics(
        metric_func, metric_class, y_true, y_pred, y_train, y_pred_benchmark
    )

    # Assertion for functions
    assert np.isclose(function_loss, true_loss), " ".join(
        [
            f"Loss function {metric_func.__name__} returned {function_loss}",
            f"loss, but {true_loss} loss expected",
        ]
    )
    # Assertion for classes
    assert np.isclose(class_loss, true_loss), " ".join(
        [
            f"Loss function {metric_class.name} returned {class_loss}",
            f"loss, but {true_loss} loss expected",
        ]
    )


@pytest.mark.parametrize("metric_func_name", LOSS_RESULTS.keys())
@pytest.mark.parametrize("n_test_case", [1, 2, 3])
def test_univariate_loss_against_expected_value(n_test_case, metric_func_name):
    metric_class = LOSS_RESULTS[metric_func_name]["class"]
    metric_func = LOSS_RESULTS[metric_func_name]["func"]
    true_loss = LOSS_RESULTS[metric_func_name][f"test_case_{n_test_case}"]
    y_true = Y_TEST_CASES[f"test_case_{n_test_case}"]["test"]
    y_train = Y_TEST_CASES[f"test_case_{n_test_case}"]["train"]

    # Use last value as naive forecast to test function
    y_pred = np.concatenate([y_train, y_true])[23:35]

    # Just using this nonsensical approach to generate  benchmark for testing
    y_pred_benchmark = 0.6 * y_pred

    function_loss, class_loss = _call_metrics(
        metric_func, metric_class, y_true, y_pred, y_train, y_pred_benchmark
    )

    # Assertion for functions
    assert np.isclose(function_loss, true_loss), " ".join(
        [
            f"Loss function {metric_func.__name__} returned {function_loss}",
            f"loss, but {true_loss} loss expected",
        ]
    )
    # Assertion for classes
    assert np.isclose(class_loss, true_loss), " ".join(
        [
            f"Loss function {metric_class.name} returned {class_loss}",
            f"loss, but {true_loss} loss expected",
        ]
    )


@pytest.mark.parametrize("metric_func_name", LOSS_RESULTS.keys())
@pytest.mark.parametrize("random_state", RANDOM_STATES)
def test_univariate_metric_function_class_equality(metric_func_name, random_state):
    metric_class = LOSS_RESULTS[metric_func_name]["class"]
    metric_func = LOSS_RESULTS[metric_func_name]["func"]

    y = _make_series(n_timepoints=75, random_state=random_state)
    y_train, y_true = y.iloc[:50], y.iloc[50:]
    y_pred = y.shift(1).iloc[50:]
    y_pred_benchmark = y.rolling(2).mean().iloc[50:]

    function_loss, class_loss = _call_metrics(
        metric_func, metric_class, y_true, y_pred, y_train, y_pred_benchmark
    )

    # Assertion for functions and class having same result
    assert np.isclose(function_loss, class_loss), " ".join(
        [
            "Expected loss function and class to return equal values,",
            f"but loss function {metric_func.__name__} returned {function_loss}",
            f"and {metric_class.name} returned {class_loss}.",
        ]
    )


@pytest.mark.parametrize("random_state", RANDOM_STATES)
@pytest.mark.parametrize("metric_func_name", LOSS_RESULTS.keys())
def test_univariate_function_output_type(metric_func_name, random_state):
    metric_func = LOSS_RESULTS[metric_func_name]["func"]
    y = _make_series(n_timepoints=75, random_state=random_state)
    y_train, y_true = y.iloc[:50], y.iloc[50:]
    y_pred = y.shift(1).iloc[50:]
    y_pred_benchmark = y.rolling(2).mean().iloc[50:]

    function_loss = metric_func(
        y_true, y_pred, y_train=y_train, y_pred_benchmark=y_pred_benchmark
    )

    is_num = is_numeric_dtype(function_loss)
    is_scalar = np.isscalar(function_loss)
    assert is_num and is_scalar, " ".join(
        ["Loss function with univariate input should return scalar number"]
    )


@pytest.mark.parametrize("metric_func_name", LOSS_RESULTS.keys())
def test_y_true_y_pred_inconsistent_n_outputs_raises_error(metric_func_name):
    metric_func = LOSS_RESULTS[metric_func_name]["func"]
    y = _make_series(n_timepoints=75, random_state=RANDOM_STATES[0])
    y_train, y_true = y.iloc[:50], y.iloc[50:]
    y_true = y_true.values  # Convert to flat NumPy array
    y_pred = y.shift(1).iloc[50:]
    y_pred = np.expand_dims(y_pred.values, 1)  # convert to 1d NumPy array
    y_pred = np.hstack([y_pred, y_pred])
    y_pred_benchmark = y.rolling(2).mean().iloc[50:]

    # Test input types
    with pytest.raises(
        ValueError, match="y_true and y_pred have different number of output"
    ):
        metric_func(y_true, y_pred, y_train=y_train, y_pred_benchmark=y_pred_benchmark)


@pytest.mark.parametrize("metric_func_name", LOSS_RESULTS.keys())
def test_y_true_y_pred_inconsistent_n_timepoints_raises_error(metric_func_name):
    metric_func = LOSS_RESULTS[metric_func_name]["func"]
    y = _make_series(n_timepoints=75, random_state=RANDOM_STATES[0])
    y_train, y_true = y.iloc[:50], y.iloc[50:]
    y_pred = y.shift(1).iloc[40:]  # y_pred has more obs
    y_pred_benchmark = y.rolling(2).mean().iloc[50:]

    # Test input types
    with pytest.raises(
        ValueError, match="Found input variables with inconsistent numbers of samples"
    ):
        metric_func(y_true, y_pred, y_train=y_train, y_pred_benchmark=y_pred_benchmark)


@pytest.mark.parametrize("metric_func_name", LOSS_RESULTS.keys())
def test_y_true_y_pred_inconsistent_n_variables_raises_error(metric_func_name):
    metric_func = LOSS_RESULTS[metric_func_name]["func"]
    y = _make_series(n_timepoints=75, random_state=RANDOM_STATES[0])
    y_train, y_true = y.iloc[:50], y.iloc[50:]
    y_true = y_true.values  # will pass as NumPy array
    y_pred = y.shift(1).iloc[50:]
    y_pred = y_pred.to_frame()
    y_pred["Second Series"] = y.shift(1).iloc[50:]
    y_pred = y_pred.values
    y_pred_benchmark = y.rolling(2).mean().iloc[50:]

    # Test input types
    with pytest.raises(
        ValueError, match="y_true and y_pred have different number of output"
    ):
        metric_func(y_true, y_pred, y_train=y_train, y_pred_benchmark=y_pred_benchmark)



# -*- coding: utf-8 -*-
""" test_experiments.py, tests the experiments code works with all current valid
classifiers.

Loops through all classifiers in the list in experiments, trains,  Builds all
"""
import os
from sktime.utils.data_io import load_from_tsfile_to_dataframe
from sktime.classification.base import classifier_list
from sktime.contrib.classification_experiments import set_classifier
from sklearn.metrics import accuracy_score
import sktime.datasets.base as sktime


class TestStats:
    def __init__(
        self,
        unit_test_acc=0.0,
        multivariate=False,
        unequal_length=False,
        missing_values=False,
    ):
        self.chinatown_acc = unit_test_acc
        self.multivariate = multivariate
        self.unequal_length = unequal_length
        self.missing_values = missing_values


# Map of classifier onto expected accuracy and expected capabilities
# If the algorithm is changed or capabilities are altered, then this needs to be
# verified and this file updated.
expected_capabilities = {
    "ProximityForest": TestStats(unit_test_acc=0.8636363636363636),
    "KNeighborsTimeSeriesClassifier": TestStats(unit_test_acc=0.8636363636363636),
    "ElasticEnsemble": TestStats(unit_test_acc=0.8636363636363636),
    "ShapeDTW": TestStats(unit_test_acc=0.8636363636363636),
    "BOSS": TestStats(unit_test_acc=0.7727272727272727),
    "ContractableBOSS": TestStats(unit_test_acc=0.8636363636363636),
    "TemporalDictionaryEnsemble": TestStats(unit_test_acc=0.8636363636363636),
    "WEASEL": TestStats(unit_test_acc=0.7272727272727273),
    "MUSE": TestStats(unit_test_acc=0.7272727272727273, multivariate=True),
    "RandomIntervalSpectralForest": TestStats(unit_test_acc=0.8636363636363636),
    "TimeSeriesForest": TestStats(unit_test_acc=0.9090909090909091),
    "CanonicalIntervalForest": TestStats(unit_test_acc=0.8636363636363636),
    "ShapeletTransformClassifier": TestStats(unit_test_acc=0.8636363636363636),
    "ROCKET": TestStats(unit_test_acc=0.9090909090909091, multivariate=True),
    "MrSEQLClassifier": TestStats(unit_test_acc=0.8636363636363636),
}


# Test that the classifiers listed in classification.base all
def test_classifiers_on_default_problem():
    path = os.path.join(sktime.MODULE, "datasets/data")
    # train_x, train_y = load_from_tsfile_to_dataframe(
    #    os.path.join(path, "UnitTest/UnitTest_TRAIN.ts")
    # )
    # test_x, test_y = load_from_tsfile_to_dataframe(
    #    os.path.join(path, "UnitTest/UnitTest_TEST.ts")
    # )
    # for name in range(0, len(classifier_list)):
    #    cls = set_classifier(name)
    # Test capabilities match expected
    # assert (
    #         cls.capabilities["multivariate"] ==
    #         expected_capabilities[name].multivariate
    # )
    # assert (
    #         cls.capabilities["unequal_length"] ==
    #         expected_capabilities[name].unequal_length
    # )
    # assert (
    #         cls.capabilities["missing_values"] ==
    #         expected_capabilities[name].missing_values
    # )
    # # Test observed accuracy matches expected accuracy
    # cls.fit(train_x, train_y)
    # preds = cls.predict(test_x)
    # ac = accuracy_score(test_y, preds)
    # assert abs(ac - expected_capabilities[name]) < 0.01


# def test_run_experiment_on_chinatown():
#    path = os.path.join(os.path.dirname(__file__), "datasets/data")#
#
#    train_x, train_y = load_from_tsfile_to_dataframe(
#        os.path.join(path, "ArrowHead/UnitTest_TRAIN.ts")
#    )
#    test_x, test_y = load_from_tsfile_to_dataframe(
#        os.path.join(path, "UnitTest/UnitTest_TEST.ts")
#    )

#    for name in range(0, len(classifier_list)):
#        cls = set_classifier(name)


# Test whether resampling is deteriministic. Resampling on fold 0 should give default
# Hard code test resampling on fold 1
# def test_stratified_resample():
#    path = os.path.join(os.path.dirname(__file__), "datasets/data")
#    train_x, train_y = load_from_tsfile_to_dataframe(
#        os.path.join(path, "UnitTest/UnitTest_TRAIN.ts")
#    )



# -*- coding: utf-8 -*-
"""Arsenal, an ensemble of ROCKET classifiers."""

__author__ = ["Matthew Middlehurst", "Oleksii Kachaiev"]
__all__ = ["Arsenal"]

import numpy as np
from joblib import Parallel, delayed
from sklearn import clone
from sklearn.ensemble._base import _set_random_states
from sklearn.linear_model import RidgeClassifierCV
from sklearn.pipeline import make_pipeline
from sklearn.utils import check_random_state
from sklearn.utils.multiclass import class_distribution

from sktime.classification.base import BaseClassifier
from sktime.transformations.panel.rocket import Rocket
from sktime.utils.validation import check_n_jobs
from sktime.utils.validation.panel import check_X
from sktime.utils.validation.panel import check_X_y


class Arsenal(BaseClassifier):
    """Ensemble of ROCKET transformers using RidgeClassifierCV base classifier.

    Allows for generation of probabilities at the expense of scalability.

    Parameters
    ----------
    num_kernels             : int, number of kernels for ROCKET transform
    (default=2,000)
    n_estimators            : int, ensemble size, optional (default=25)
    n_jobs                  : int, the number of jobs to run in parallel for both `fit`
    and `predict`. ``-1`` means using all processors
    random_state            : int or None, seed for random, integer,
    optional (default to no seed)

    Attributes
    ----------
    estimators_             : array of individual classifiers
    weights                 : weight of each classifier in the ensemble
    weight_sum              : sum of all weights
    n_classes               : extracted from the data

    Notes
    -----
    @article{dempster_etal_2019,
      author  = {Dempster, Angus and Petitjean, Francois and Webb,
      Geoffrey I},
      title   = {ROCKET: Exceptionally fast and accurate time series
      classification using random convolutional kernels},
      year    = {2019},
      journal = {arXiv:1910.13051}
    }

    Java version
    https://github.com/uea-machine-learning/tsml/blob/master/src/main/java/
    tsml/classifiers/shapelet_based/Arsenal.java

    """

    # Capability tags
    capabilities = {
        "multivariate": True,
        "unequal_length": False,
        "missing_values": False,
        "train_estimate": False,
        "contractable": False,
    }

    def __init__(
        self,
        num_kernels=2000,
        n_estimators=25,
        n_jobs=1,
        random_state=None,
    ):
        self.num_kernels = num_kernels
        self.n_estimators = n_estimators
        self.n_jobs = n_jobs
        self.random_state = random_state

        self.estimators_ = []
        self.weights = []
        self.weight_sum = 0

        self.n_classes = 0
        self.classes_ = []
        self.class_dictionary = {}

        super(Arsenal, self).__init__()

    def fit(self, X, y):
        """Build an ensemble ROCKET transformer and RidgeClassifierCV classifier.

        Parameters
        ----------
        X : nested pandas DataFrame of shape [n_instances, 1]
            Nested dataframe with univariate time-series in cells.
        y : array-like, shape = [n_instances] The class labels.

        Returns
        -------
        self : object
        """
        X, y = check_X_y(X, y)
        n_jobs = check_n_jobs(self.n_jobs)

        self.n_classes = np.unique(y).shape[0]
        self.classes_ = class_distribution(np.asarray(y).reshape(-1, 1))[0][0]
        for index, classVal in enumerate(self.classes_):
            self.class_dictionary[classVal] = index

        base_estimator = _make_estimator(self.num_kernels, self.random_state)
        self.estimators_ = Parallel(n_jobs=n_jobs)(
            delayed(_fit_estimator)(
                _clone_estimator(base_estimator, self.random_state, i), X, y
            )
            for i in range(self.n_estimators)
        )

        self.weights = []
        self.weight_sum = 0
        for rocket_pipeline in self.estimators_:
            weight = rocket_pipeline.steps[1][1].best_score_
            self.weights.append(weight)
            self.weight_sum += weight

        self._is_fitted = True
        return self

    def predict(self, X):
        """Find predictions for all cases in X. Built on top of predict_proba.

        Parameters
        ----------
        X : The training input samples. array-like or pandas data frame.
        If a Pandas data frame is passed, a check is performed that it only
        has one column.
        If not, an exception is thrown, since this classifier does not yet have
        multivariate capability.

        Returns
        -------
        output : array of shape = [n_test_instances]
        """
        rng = check_random_state(self.random_state)
        return np.array(
            [
                self.classes_[int(rng.choice(np.flatnonzero(prob == prob.max())))]
                for prob in self.predict_proba(X)
            ]
        )

    def predict_proba(self, X):
        """Find probability estimates for each class for all cases in X.

        Parameters
        ----------
        X : The training input samples. array-like or sparse matrix of shape
        = [n_test_instances, series_length]
            If a Pandas data frame is passed (sktime format) a check is
            performed that it only has one column.
            If not, an exception is thrown, since this classifier does not
            yet have
            multivariate capability.

        Returns
        -------
        output : array of shape = [n_test_instances, num_classes] of
        probabilities
        """
        self.check_is_fitted()
        X = check_X(X, coerce_to_numpy=True)

        sums = np.zeros((X.shape[0], self.n_classes))

        for n, clf in enumerate(self.estimators_):
            preds = clf.predict(X)
            for i in range(0, X.shape[0]):
                sums[i, self.class_dictionary[preds[i]]] += self.weights[n]

        return np.around(sums / (np.ones(self.n_classes) * self.weight_sum), 8)

    def _get_train_probs(self, X):
        return 0


def _fit_estimator(estimator, X, y):
    return estimator.fit(X, y)


def _make_estimator(num_kernels, random_state):
    return make_pipeline(
        Rocket(num_kernels=num_kernels, random_state=random_state),
        RidgeClassifierCV(alphas=np.logspace(-3, 3, 10), normalize=True),
    )


def _clone_estimator(base_estimator, random_state=None, idx=0):
    estimator = clone(base_estimator)

    if random_state is not None:
        random_state = 255 if random_state == 0 else random_state
        _set_random_states(estimator, random_state * 37 * (idx + 1))

    return estimator



# -*- coding: utf-8 -*-
"""Base classes for clustering."""
__author__ = ["Christopher Holder", "Tony Bagnall"]
__all__ = [
    "BaseClusterer",
    "BaseClusterCenterInitializer",
    "BaseClusterAverage",
    "BaseApproximate",
    "DataFrame",
    "NumpyArray",
    "NumpyOrDF",
    "NumpyRandomState",
    "CenterCalculatorFunc",
]

import pandas as pd
import numpy as np
from sktime.base import BaseEstimator
from sktime.utils.data_processing import from_nested_to_2d_array
from typing import Union, Callable

DataFrame = pd.DataFrame
NumpyArray = np.ndarray
NumpyOrDF = Union[DataFrame, NumpyArray]
NumpyRandomState = Union[np.random.RandomState, int]
CenterCalculatorFunc = Callable[[NumpyArray], NumpyArray]


class BaseClusterer(BaseEstimator):
    """Base Clusterer"""

    def __init__(self):
        super(BaseClusterer, self).__init__()

    def fit(self, X: NumpyOrDF, y=None):
        """
        Fit the clustering algorithm on the dataset X

        Parameters
        ----------
        X: 2D np.array with shape (n_instances, n_timepoints)
           or pd.DataFrame in nested format
            panel of univariate time series to train the clustering model on

        y: ignored, exists for API consistency reasons

        Returns
        -------
        reference to self
        """
        if isinstance(X, pd.DataFrame):
            X = from_nested_to_2d_array(X, return_numpy=True)

        self._check_params(X)
        self._fit(X)

        self._is_fitted = True
        return self

    def predict(self, X: NumpyOrDF, y=None) -> NumpyArray:
        """
        Return cluster center index for data samples.

        Parameters
        ----------
        X: 2D np.array with shape (n_instances, n_timepoints)
           or pd.DataFrame in nested format
            panel of time series to cluster

        y: ignored, exists for API consistency reasons

        Returns
        -------
        Numpy_Array: 1D np.array of length n_instances
            Index of the cluster each sample belongs to
        """
        self.check_is_fitted()

        if isinstance(X, pd.DataFrame):
            X = from_nested_to_2d_array(X, return_numpy=True)

        self._check_params(X)
        return self._predict(X)

    def _fit(self, X: NumpyArray):
        """
        Fit the clustering algorithm on the dataset X

            core logic

        Parameters
        ----------
        X: 2D np.array with shape (n_instances, n_timepoints)
            panel of univariate time series to train the clustering model on

        Returns
        -------
        reference to self
        """
        raise NotImplementedError("abstract method")

    def _predict(self, X: NumpyArray) -> NumpyArray:
        """
        Return cluster center index for data samples.

            core logic

        Parameters
        ----------
        X: 2D np.array with shape (n_instances, n_timepoints)
            panel of univariate time series to cluster

        Returns
        -------
        Numpy_Array: 1D np.array of length n_instances
            Index of the cluster each sample belongs to
        """
        raise NotImplementedError("abstract method")

    def _check_params(self, X: NumpyArray):
        """
        Custom input checking, should raise errors

        Parameters
        ----------
        X: 2D np.array with shape (n_instances, n_timepoints)
        """
        return

    def fit_predict(self, X: NumpyOrDF, y=None) -> NumpyArray:
        """
        clusters time series and returns cluster labels

        Parameters
        ----------
        X: 2D np.array with shape (n_instances, n_timepoints)
           or pd.DataFrame in nested format
            panel of univariate time series to cluster

        y: ignored, exists for API consistency reasons

        Returns
        -------
        Numpy_Array: 1D np.array of length n_instances
            Index of the cluster each sample belongs to
        """
        self.fit(X)
        return self.predict(X)

    def get_fitted_params(self):
        """
        Retrieves fitted parameters of cluster model

        returns
        ----------
        param_dict: dictionary of fitted parameters
        """
        self.check_is_fitted()

        return self._get_fitted_params()

    def _get_fitted_params(self):
        """
        Retrieves fitted parameters of cluster model

            core logic

        returns
        ----------
        param_dict: dictionary of fitted parameters
        """
        raise NotImplementedError


class BaseClusterCenterInitializer:
    """Base Cluster Center Initializer

    Parameters
    ----------
    data_set: Numpy_Array
        Numpy_Array that is the dataset to calculate the centers from

    n_centers: int
        Number of centers to be created

    center_calculator_func: CenterCalculatorFunc, default = None
        Function that is used to calculate new centers

    random_state: NumpyRandomState, default = None
        Generator used to initialise the centers.
    """

    def __init__(
        self,
        data_set: NumpyArray,
        n_centers: int,
        center_calculator_func: CenterCalculatorFunc = None,
        random_state: NumpyRandomState = None,
    ):
        self.data_set: data_set = data_set
        self.n_centers: int = n_centers
        self.center_calculator_func: CenterCalculatorFunc = center_calculator_func
        self.random_state: NumpyRandomState = random_state

    def initialize_centers(self) -> NumpyArray:
        """
        Method used to initialise centers

        Returns
        -------
        Data_Frame
            sktime dataframe containing the starting
            centers
        """
        raise NotImplementedError("abstract method")


class BaseClusterAverage:
    """Base Cluster Average

    Parameters
    ----------
    series: Numpy_Array
        Set of series to generate a average from
    """

    def __init__(self, series: NumpyArray, n_iterations: int = 10):
        self.n_iterations = n_iterations
        self.series = series

    def average(self) -> NumpyArray:
        """
        Method called to find the average for a distance metric

        Returns
        -------
        Numpy_Array
            Created array denoting the average

        n_iterations: int
            Number of iterations to refine the average
        """
        raise NotImplementedError("abstract method")


class BaseApproximate:
    """Base Approximate

    Parameters
    ----------
    series: Numpy_Array
        series to perform approximation on

    """

    def __init__(self, series: NumpyArray):
        self.series = series

    def approximate(self) -> int:
        """
        Method called to get the approximation

        Returns
        -------
        int
            Index position of the approximation in the series
        """
        raise NotImplementedError("abstract method")



#!/usr/bin/env python3 -u
# -*- coding: utf-8 -*-
import pytest

from sktime.utils.validation import check_window_length


@pytest.mark.parametrize(
    "window_length, n_timepoints, expected",
    [
        (0.2, 33, 7),
        (43, 23, 43),
        (33, 1, 33),
        (33, None, 33),
        (None, 19, None),
        (None, None, None),
        (67, 0.3, 67),  # bad arg
    ],
)
def test_check_window_length(window_length, n_timepoints, expected):
    assert check_window_length(window_length, n_timepoints) == expected


@pytest.mark.parametrize(
    "window_length, n_timepoints",
    [
        ("string", 34),
        ("string", "string"),
        (6.2, 33),
        (-5, 34),
        (-0.5, 15),
        (6.1, 0.3),
        (0.3, 0.1),
        (-2.4, 10),
        (0.2, None),
    ],
)
def test_window_length_bad_arg(window_length, n_timepoints):
    with pytest.raises(ValueError):
        check_window_length(window_length, n_timepoints)



# -*- coding: utf-8 -*-
"""
module exports for dist_kernels module
"""

from sktime.dists_kernels._base import (
    BasePairwiseTransformer,
    BasePairwiseTransformerPanel,
)
from sktime.dists_kernels.compose_tab_to_panel import AggrDist
from sktime.dists_kernels.scipy_dist import ScipyDist

__all__ = [
    "BasePairwiseTransformer",
    "BasePairwiseTransformerPanel",
    "AggrDist",
    "ScipyDist",
]
